"""update stagedresource chid_diff_statuses to booleans

Revision ID: 9cad5a5c438c
Revises: d9064e71f69d
Create Date: 2024-08-27 07:45:33.495919

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy import text
from sqlalchemy.engine import Connection

# revision identifiers, used by Alembic.
revision = "9cad5a5c438c"
down_revision = "d9064e71f69d"
branch_labels = None
depends_on = None


def upgrade():
    bind: Connection = op.get_bind()
    # upgrade query sets child_diff_statuses JSON keys to have a value of 'true' if they currently have a value > 0;
    # otherwise, the key is not added.
    # this is done for each diff status, which are the possible keys in child_diff_statuses:
    # - monitored
    # - addition
    # - removal
    # - muted
    # - classification_addition
    # - classification_update
    upgrade_query = text(
        """
update
   stagedresource
set
   child_diff_statuses = '{}' || jsonb_strip_nulls(jsonb_build_object('monitored',
   case
      when
         jsonb_typeof(child_diff_statuses -> 'monitored') = 'number'
         and
         (
            child_diff_statuses -> 'monitored'
         )
         ::integer > 0
      then
         true
      else
         null
   end
)) || jsonb_strip_nulls(jsonb_build_object('addition',
   case
      when
         jsonb_typeof(child_diff_statuses -> 'addition') = 'number'
         and
         (
            child_diff_statuses -> 'addition'
         )
         ::integer > 0
      then
         true
      else
         null
   end
)) || jsonb_strip_nulls(jsonb_build_object('removal',
   case
      when
         jsonb_typeof(child_diff_statuses -> 'removal') = 'number'
         and
         (
            child_diff_statuses -> 'removal'
         )
         ::integer > 0
      then
         true
      else
         null
   end
)) || jsonb_strip_nulls(jsonb_build_object('muted',
   case
      when
         jsonb_typeof(child_diff_statuses -> 'muted') = 'number'
         and
         (
            child_diff_statuses -> 'muted'
         )
         ::integer > 0
      then
         true
      else
         null
   end
)) || jsonb_strip_nulls(jsonb_build_object('classification_addition',
   case
      when
         jsonb_typeof(child_diff_statuses -> 'classification_addition') = 'number'
         and
         (
            child_diff_statuses -> 'classification_addition'
         )
         ::integer > 0
      then
         true
      else
         null
   end
)) || jsonb_strip_nulls(jsonb_build_object('classification_update',
   case
      when
         jsonb_typeof(child_diff_statuses -> 'classification_update') = 'number'
         and
         (
            child_diff_statuses -> 'classification_update'
         )
         ::integer > 0
      then
         true
      else
         null
   end
))
"""
    )
    bind.execute(upgrade_query)


def downgrade():
    bind: Connection = op.get_bind()
    # downgrade query sets child_diff_statuses JSON keys to have a value of '1' if they currently have a value of true;
    # otherwise, the key is not added.
    # this is done for each diff status, which are the possible keys in child_diff_statuses:
    # - monitored
    # - addition
    # - removal
    # - muted
    # - classification_addition
    # - classification_update
    downgrade_query = text(
        """
update
   stagedresource
set
   child_diff_statuses = '{}' || jsonb_strip_nulls(jsonb_build_object('monitored',
   case
      when
         jsonb_typeof(child_diff_statuses -> 'monitored') = 'boolean'
         and
         (
            child_diff_statuses -> 'monitored'
         )
         ::boolean
      then
         1
      else
         null
   end
)) || jsonb_strip_nulls(jsonb_build_object('addition',
   case
      when
         jsonb_typeof(child_diff_statuses -> 'addition') = 'boolean'
         and
         (
            child_diff_statuses -> 'addition'
         )
         ::boolean
      then
         1
      else
         null
   end
)) || jsonb_strip_nulls(jsonb_build_object('removal',
   case
      when
         jsonb_typeof(child_diff_statuses -> 'removal') = 'boolean'
         and
         (
            child_diff_statuses -> 'removal'
         )
         ::boolean
      then
         1
      else
         null
   end
)) || jsonb_strip_nulls(jsonb_build_object('muted',
   case
      when
         jsonb_typeof(child_diff_statuses -> 'muted') = 'boolean'
         and
         (
            child_diff_statuses -> 'muted'
         )
         ::boolean
      then
         1
      else
         null
   end
)) || jsonb_strip_nulls(jsonb_build_object('classification_addition',
   case
      when
         jsonb_typeof(child_diff_statuses -> 'classification_addition') = 'boolean'
         and
         (
            child_diff_statuses -> 'classification_addition'
         )
         ::boolean
      then
         1
      else
         null
   end
)) || jsonb_strip_nulls(jsonb_build_object('classification_update',
   case
      when
         jsonb_typeof(child_diff_statuses -> 'classification_update') = 'boolean'
         and
         (
            child_diff_statuses -> 'classification_update'
         )
         ::boolean
      then
         1
      else
         null
   end
))
"""
    )
    bind.execute(downgrade_query)



from enum import Enum
from typing import Dict, List, Optional

from fides.common.api.scope_registry import (
    CLI_OBJECTS_READ,
    CLIENT_READ,
    CONNECTION_READ,
    CONNECTION_TYPE_READ,
    CONNECTOR_TEMPLATE_REGISTER,
    CONSENT_READ,
    CONSENT_SETTINGS_READ,
    CTL_DATASET_READ,
    CTL_POLICY_READ,
    DATA_CATEGORY_READ,
    DATA_SUBJECT_READ,
    DATA_USE_READ,
    DATASET_READ,
    EVALUATION_READ,
    MASKING_EXEC,
    MASKING_READ,
    MESSAGING_CREATE_OR_UPDATE,
    MESSAGING_DELETE,
    MESSAGING_READ,
    ORGANIZATION_READ,
    POLICY_READ,
    PRIVACY_EXPERIENCE_READ,
    PRIVACY_NOTICE_READ,
    PRIVACY_REQUEST_CALLBACK_RESUME,
    PRIVACY_REQUEST_DELETE,
    PRIVACY_REQUEST_NOTIFICATIONS_CREATE_OR_UPDATE,
    PRIVACY_REQUEST_NOTIFICATIONS_READ,
    PRIVACY_REQUEST_READ,
    PRIVACY_REQUEST_REVIEW,
    PRIVACY_REQUEST_UPLOAD_DATA,
    PRIVACY_REQUEST_VIEW_DATA,
    RULE_READ,
    SAAS_CONFIG_READ,
    SCOPE_READ,
    SCOPE_REGISTRY,
    STORAGE_CREATE_OR_UPDATE,
    STORAGE_DELETE,
    STORAGE_READ,
    SYSTEM_MANAGER_READ,
    SYSTEM_READ,
    USER_PERMISSION_ASSIGN_OWNERS,
    USER_READ,
    WEBHOOK_READ,
)

APPROVER = "approver"
CONTRIBUTOR = "contributor"
OWNER = "owner"
VIEWER = "viewer"
VIEWER_AND_APPROVER = "viewer_and_approver"


class RoleRegistryEnum(Enum):
    """Enum of available roles

    Owner - Full admin
    Viewer - Can view everything
    Approver - Limited viewer but can approve Privacy Requests
    Viewer + Approver = Full View and can approve Privacy Requests
    Contributor - Can't configure storage and messaging
    """

    owner = OWNER
    viewer_approver = VIEWER_AND_APPROVER
    viewer = VIEWER
    approver = APPROVER
    contributor = CONTRIBUTOR


approver_scopes = [
    PRIVACY_REQUEST_REVIEW,
    PRIVACY_REQUEST_READ,
    PRIVACY_REQUEST_CALLBACK_RESUME,
    PRIVACY_REQUEST_UPLOAD_DATA,
    PRIVACY_REQUEST_VIEW_DATA,
    PRIVACY_REQUEST_DELETE,
    USER_READ,  # allows approver to view user management table and update their own password
]


viewer_scopes = [  # Intentionally omitted USER_PERMISSION_READ and PRIVACY_REQUEST_READ
    CLI_OBJECTS_READ,
    CLIENT_READ,
    CONNECTION_READ,
    CONSENT_READ,
    CONSENT_SETTINGS_READ,
    CONNECTION_TYPE_READ,
    CTL_DATASET_READ,
    DATA_CATEGORY_READ,
    CTL_POLICY_READ,
    DATASET_READ,
    DATA_SUBJECT_READ,
    DATA_USE_READ,
    EVALUATION_READ,
    MASKING_EXEC,
    MASKING_READ,
    ORGANIZATION_READ,
    POLICY_READ,
    PRIVACY_EXPERIENCE_READ,
    PRIVACY_NOTICE_READ,
    PRIVACY_REQUEST_NOTIFICATIONS_READ,
    RULE_READ,
    SCOPE_READ,
    STORAGE_READ,
    SYSTEM_READ,
    MESSAGING_READ,
    WEBHOOK_READ,
    SYSTEM_MANAGER_READ,
    SAAS_CONFIG_READ,
    USER_READ,
]

not_contributor_scopes = [
    CONNECTOR_TEMPLATE_REGISTER,
    STORAGE_CREATE_OR_UPDATE,
    STORAGE_DELETE,
    MESSAGING_CREATE_OR_UPDATE,
    MESSAGING_DELETE,
    PRIVACY_REQUEST_NOTIFICATIONS_CREATE_OR_UPDATE,
    USER_PERMISSION_ASSIGN_OWNERS,
]

ROLES_TO_SCOPES_MAPPING: Dict[str, List] = {
    OWNER: sorted(SCOPE_REGISTRY),
    VIEWER_AND_APPROVER: sorted(list(set(viewer_scopes + approver_scopes))),
    VIEWER: sorted(viewer_scopes),
    APPROVER: sorted(approver_scopes),
    CONTRIBUTOR: sorted(list(set(SCOPE_REGISTRY) - set(not_contributor_scopes))),
}


def get_scopes_from_roles(roles: Optional[List[str]]) -> List[str]:
    """Return a list of all the scopes the user has via their role(s)"""
    if not roles:
        return []

    scope_list: List[str] = []
    for role in roles:
        scope_list += ROLES_TO_SCOPES_MAPPING.get(role, [])
    return [*set(scope_list)]



from rich_click import secho


class LocalModeException(Exception):
    """Raise an error when `local_mode` is enabled with an incompatible command."""

    def __init__(self, command: str):
        self.command = command
        self.message = f"> Command: `{self.command}` not compatible with `local_mode`!"
        secho(self.message, fg="red")
        raise SystemExit(1)



"""Module that adds functionality for generating or scanning datasets."""

from typing import Dict, List, Optional, Tuple

import sqlalchemy
from fideslang import manifests
from fideslang.models import Dataset, DatasetCollection, DatasetField
from fideslang.validation import AnyHttpUrlString, FidesKey
from joblib import Parallel, delayed
from sqlalchemy.engine import Engine
from sqlalchemy.sql import text

from fides.common.utils import echo_green, echo_red
from fides.connectors.aws import (
    create_dynamodb_dataset,
    describe_dynamo_tables,
    get_aws_client,
    get_dynamo_tables,
)
from fides.connectors.bigquery import get_bigquery_engine
from fides.connectors.models import AWSConfig, BigQueryConfig
from fides.core.api_helpers import list_server_resources
from fides.core.parse import parse
from fides.core.utils import check_fides_key, generate_unique_fides_key, get_db_engine

SCHEMA_EXCLUSION = {
    "postgresql": ["information_schema"],
    "mysql": ["mysql", "performance_schema", "sys", "information_schema"],
    "mssql": ["INFORMATION_SCHEMA", "guest", "sys"],
    "snowflake": ["INFORMATION_SCHEMA"],
    "redshift": ["information_schema"],
}


def get_all_server_datasets(
    url: AnyHttpUrlString, headers: Dict[str, str], exclude_datasets: List[Dataset]
) -> List[Dataset]:
    """
    Get a list of all of the Datasets that exist on the server. Excludes any datasets
    provided in exclude_datasets
    """
    exclude_dataset_keys = [dataset.fides_key for dataset in exclude_datasets]
    raw_dataset_list = (
        list_server_resources(
            url=str(url),
            resource_type="dataset",
            exclude_keys=[str(x) for x in exclude_dataset_keys],
            headers=headers,
        )
        or []
    )
    dataset_list = [Dataset.model_validate(dataset) for dataset in raw_dataset_list]

    return dataset_list


def include_dataset_schema(schema: str, database_type: str) -> bool:
    """
    Returns whether a schema should be included based on database type and schema name
    """
    schema_exclusion_list = SCHEMA_EXCLUSION.get(database_type, [])
    if database_type == "mssql":
        return schema not in schema_exclusion_list and not schema.startswith("db_")
    return schema not in schema_exclusion_list


def get_db_schemas(
    engine: Engine,
) -> Dict[str, Dict[str, List[str]]]:
    """
    Extract the schema, table and column names from a database given a sqlalchemy engine
    """
    if engine.dialect.name != "snowflake":
        inspector = sqlalchemy.inspect(engine)
        db_schemas: Dict[str, Dict[str, List]] = {}
        for schema in inspector.get_schema_names():
            if include_dataset_schema(schema=schema, database_type=engine.dialect.name):
                db_schemas[schema] = {}
                for table in inspector.get_table_names(schema=schema):
                    db_schemas[schema][table] = [
                        column["name"]
                        for column in inspector.get_columns(table, schema=schema)
                    ]
    else:
        db_schemas = get_snowflake_schemas(engine=engine)
    return db_schemas


def create_db_datasets(db_schemas: Dict[str, Dict[str, List[str]]]) -> List[Dataset]:
    """
    Returns a list of dataset objects given a database schema
    mapping.
    """
    datasets = [
        create_db_dataset(schema_name=schema_name, db_tables=db_tables)
        for schema_name, db_tables in db_schemas.items()
    ]
    return datasets


def create_db_dataset(schema_name: str, db_tables: Dict[str, List[str]]) -> Dataset:
    """
    Returns a Dataset object of a database schema given a schema name
    and table definitions.

    Placeholder values are use where needed.
    """
    sanitized_fides_key = check_fides_key(schema_name)
    dataset = Dataset(
        fides_key=sanitized_fides_key,
        name=schema_name,
        description=f"Fides Generated Description for Schema: {schema_name}",
        data_categories=[],
        collections=[
            DatasetCollection(
                name=table_name,
                description=f"Fides Generated Description for Table: {table_name}",
                data_categories=[],
                fields=[
                    DatasetField(
                        name=column,
                        description=f"Fides Generated Description for Column: {column}",
                        data_categories=[],
                    )
                    for column in table
                ],
            )
            for table_name, table in db_tables.items()
        ],
    )
    return dataset


def make_dataset_key_unique(
    dataset: Dataset, database_host: str, database_name: str
) -> Dataset:
    """
    Ensure the dataset has a repeatable unique ID appended to the end
    to avoid naming collisions.
    """

    dataset.fides_key = str(  # type: ignore
        FidesKey(
            generate_unique_fides_key(dataset.fides_key, database_host, database_name)
        )
    )
    dataset.meta = {"database_host": database_host, "database_name": database_name}
    return dataset


def find_uncategorized_dataset_fields(
    existing_dataset: Optional[Dataset], source_dataset: Dataset
) -> Tuple[List[str], int]:
    """
    Iterates all fields in a source dataset and finds which are not
    defined or categorized in an existing dataset. Returns fields that
    are not categorized or defined, and also the total fields iterated.
    """
    uncategorized_fields = []
    total_field_count = 0

    for source_dataset_collection in source_dataset.collections:
        existing_dataset_collection = (
            next(
                (
                    existing_dataset_collection
                    for existing_dataset_collection in existing_dataset.collections
                    if existing_dataset_collection.name
                    == source_dataset_collection.name
                ),
                None,
            )
            if existing_dataset
            else None
        )

        for db_dataset_field in source_dataset_collection.fields:
            total_field_count += 1
            field_uncategorized = (
                all(
                    field.name != db_dataset_field.name or not field.data_categories
                    for field in existing_dataset_collection.fields
                )
                if existing_dataset_collection
                else True
            )

            if field_uncategorized:
                uncategorized_fields.append(
                    f"{source_dataset.name}.{source_dataset_collection.name}.{db_dataset_field.name}"
                )
    return uncategorized_fields, total_field_count


def find_all_uncategorized_dataset_fields(
    existing_datasets: List[Dataset], source_datasets: List[Dataset]
) -> Tuple[List[str], int]:
    """
    Finds all uncategorized fields given a list of source datasets
    and existing datasets to compare against.
    """
    uncategorized_fields = []
    total_field_count = 0
    for source_dataset in source_datasets:
        existing_dataset = next(
            (
                existing_dataset
                for existing_dataset in existing_datasets
                if existing_dataset.name == source_dataset.name
            ),
            None,
        )

        (
            current_uncategorized_keys,
            current_field_count,
        ) = find_uncategorized_dataset_fields(
            existing_dataset=existing_dataset, source_dataset=source_dataset
        )
        total_field_count += current_field_count
        uncategorized_fields += current_uncategorized_keys

    return uncategorized_fields, total_field_count


def print_dataset_db_scan_result(
    datasets: List[Dataset],
    uncategorized_fields: List[str],
    coverage_percent: int,
    coverage_threshold: int,
) -> None:
    """
    Prints uncategorized fields and raises an exception if coverage
    is lower than provided threshold.
    """
    dataset_names: List[str] = [dataset.name or "" for dataset in datasets]
    output: str = "Successfully scanned the following datasets:\n"
    output += "\t{}\n".format("\n\t".join(dataset_names))
    echo_green(output)

    if uncategorized_fields:
        uncategorized_output = (
            "The following fields are missing data category annotations:\n"
        )
        uncategorized_output += "\t{}\n".format("\n\t".join(uncategorized_fields))
        echo_red(uncategorized_output)
    annotation_output = "Annotation coverage: {}%".format(coverage_percent)
    if coverage_percent < coverage_threshold:
        echo_red(annotation_output)
        raise SystemExit(1)
    echo_green(annotation_output)


def scan_dataset_db(
    connection_string: str,
    manifest_dir: Optional[str],
    coverage_threshold: int,
    url: AnyHttpUrlString,
    headers: Dict[str, str],
    local: bool = False,
) -> None:
    """
    Given a database connection string, fetches collections
    and fields and compares them to existing datasets and prioritizes
    datasets in a local manifest (if one is provided).
    """

    if manifest_dir:
        manifest_datasets = parse(manifest_dir).dataset or []
    else:
        manifest_datasets = []

    if not local:
        server_datasets = (
            get_all_server_datasets(
                url=url, headers=headers, exclude_datasets=manifest_datasets
            )
            or []
        )
    else:
        server_datasets = []

    all_datasets = manifest_datasets + server_datasets

    dataset_keys = [dataset.fides_key for dataset in all_datasets]
    echo_green(
        "Loaded the following dataset manifests:\n\t{}".format(
            "\n\t".join(dataset_keys)
        )
    )

    # Generate the collections and fields for the target database
    db_datasets = generate_db_datasets(connection_string=connection_string)
    uncategorized_fields, db_field_count = find_all_uncategorized_dataset_fields(
        existing_datasets=all_datasets,
        source_datasets=db_datasets,
    )
    if db_field_count < 1:
        echo_red("Database did not contain any dataset fields to evaluate coverage")
        raise SystemExit(1)

    coverage_percent = int(
        ((db_field_count - len(uncategorized_fields)) / db_field_count) * 100
    )
    print_dataset_db_scan_result(
        datasets=db_datasets,
        uncategorized_fields=uncategorized_fields,
        coverage_percent=coverage_percent,
        coverage_threshold=coverage_threshold,
    )


def generate_db_datasets(connection_string: str) -> List[Dataset]:
    """
    Given a database connection string, extract all tables/fields from it
    and generate corresponding datasets.
    """
    db_engine = get_db_engine(connection_string)
    db_schemas = get_db_schemas(engine=db_engine)
    db_datasets = create_db_datasets(db_schemas=db_schemas)
    unique_db_datasets = [
        make_dataset_key_unique(dataset, db_engine.url.host, db_engine.url.database)
        for dataset in db_datasets
    ]
    return unique_db_datasets


def write_dataset_manifest(
    file_name: str, include_null: bool, datasets: List[Dataset]
) -> None:
    """
    Given a list of datasets, writes a manifest file with the given datasets.

    An optional flag can be passed to include null fields.
    """
    manifests.write_manifest(
        file_name,
        [i.model_dump(exclude_none=not include_null) for i in datasets],
        "dataset",
    )
    echo_green(f"Generated dataset manifest written to {file_name}")


def generate_dataset_db(
    connection_string: str, file_name: str, include_null: bool
) -> str:
    """
    Given a database connection string, extract all tables/fields from it
    and write out a boilerplate dataset manifest, excluding optional null attributes.
    """
    db_datasets = generate_db_datasets(connection_string=connection_string)
    write_dataset_manifest(
        file_name=file_name, include_null=include_null, datasets=db_datasets
    )
    return file_name


def generate_bigquery_datasets(bigquery_config: BigQueryConfig) -> List[Dataset]:
    """
    Given a BigQuery config, extract all tables/fields and generate corresponding datasets.
    """
    bigquery_engine = get_bigquery_engine(bigquery_config)
    bigquery_schemas = get_db_schemas(engine=bigquery_engine)
    bigquery_datasets = create_db_datasets(db_schemas=bigquery_schemas)
    unique_bigquery_datasets = [
        make_dataset_key_unique(
            dataset, bigquery_engine.url.host, bigquery_engine.url.database
        )
        for dataset in bigquery_datasets
    ]
    return unique_bigquery_datasets


def generate_dynamo_db_datasets(
    aws_config: Optional[AWSConfig], single_dataset: bool
) -> List[Dataset]:
    """
    Given an AWS config, extract all DynamoDB tables/fields and generate corresponding datasets.
    """
    client = get_aws_client(service="dynamodb", aws_config=aws_config)
    dynamo_tables = get_dynamo_tables(client)
    described_dynamo_tables = describe_dynamo_tables(client, dynamo_tables)
    dynamo_dataset = create_dynamodb_dataset(
        described_dynamo_tables, single_dataset=single_dataset
    )
    return dynamo_dataset


def get_snowflake_schemas(
    engine: Engine,
) -> Dict[str, Dict[str, List[str]]]:
    """
    Returns Datasets that match the case-sensitivity that may be
    required by Snowflake. Iterates through each schema.table.column
    the logged in user has access to.

    This is currently required because of the inferred casing of Snowflake,
    which defaults to upper-case. Anything else must be double-quoted, however
    `snowflake-sqlalchemy` does not account for this in a flexible manner in
    it's implementation of the `inspect()` method, forcing everything to what
    is deemed to be normalized (i.e. lower-case).

    The following code maintains casing as defined in Snowflake which combines
    well with our DSR implementation in always using double-quoted query syntax.

    It may be worthwhile for us to invest some time in resolving the core issue
    and being able to fall back to using the connector.

    Reference: https://github.com/snowflakedb/snowflake-sqlalchemy/issues/157

    Due to performance issues resulting in FastAPI timeouts, parallelization is
    now used. The default number of threads to be used is 4 to cover most existing
    use cases, bumping to 8 when the number of tables is over 250.
    """
    schema_cursor = engine.execute(text("SHOW SCHEMAS"))
    db_schemas = [row[1] for row in schema_cursor]
    metadata: Dict[str, Dict[str, List]] = {}
    for schema in db_schemas:
        if include_dataset_schema(schema=schema, database_type=engine.dialect.name):
            metadata[schema] = {}
            table_cursor = engine.execute(text(f'SHOW TABLES IN "{schema}"'))
            db_tables = [row[1] for row in table_cursor]
            number_of_threads = 8 if len(db_tables) > 250 else 4
            fields = Parallel(n_jobs=number_of_threads, backend="threading")(
                delayed(get_snowflake_table_fields)(engine, schema, table)
                for table in db_tables
            )
            for table, columns in fields:
                metadata[schema][table] = columns
    return metadata


def get_snowflake_table_fields(
    engine: Engine, schema: str, table: str
) -> Tuple[str, List]:
    """
    Returns fields for a Snowflake table, ideally in parallel. Part of the
    workaround strategy to improve performance when generating a Snowflake schema
    """
    column_cursor = engine.execute(text(f'SHOW COLUMNS IN "{schema}"."{table}"'))
    columns = [row[2] for row in column_cursor]
    return table, columns



from datetime import datetime
from typing import List

from sqlalchemy import Column, DateTime, String
from sqlalchemy.dialects.postgresql import ARRAY
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.mutable import MutableList
from sqlalchemy.orm import Session

from fides.api.db.base_class import Base


class SystemCompassSync(Base):
    """A record of systems updated via Compass sync"""

    @declared_attr
    def __tablename__(self) -> str:
        return "system_compass_sync"

    sync_started_at = Column(DateTime(timezone=True))
    sync_completed_at = Column(DateTime(timezone=True))

    updated_systems = Column(MutableList.as_mutable(ARRAY(String)))

    @classmethod
    def start_system_sync(cls, db: Session) -> "SystemCompassSync":
        return SystemCompassSync(sync_started_at=datetime.utcnow()).save(db=db)  # type: ignore[return-value]

    def finish_system_sync(
        self, db: Session, updated_systems: List[str]
    ) -> "SystemCompassSync":
        self.updated_systems = updated_systems
        self.sync_completed_at = datetime.utcnow()
        return self.save(db=db)  # type: ignore[return-value]



"""save privacy preferences on device id

Revision ID: 451684a726a5
Revises: 48d9caacebd4
Create Date: 2023-04-23 16:06:19.788074

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op

# revision identifiers, used by Alembic.
revision = "451684a726a5"
down_revision = "48d9caacebd4"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "currentprivacypreference",
        sa.Column("fides_user_device_provided_identity_id", sa.String(), nullable=True),
    )
    op.create_unique_constraint(
        "fides_user_device_identity_privacy_notice",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id", "privacy_notice_id"],
    )
    op.create_index(
        op.f("ix_currentprivacypreference_fides_user_device_provided_identity_id"),
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id"],
        unique=False,
    )
    op.create_foreign_key(
        "currentprivacypreference_fides_user_device_provided_identi_fkey",
        "currentprivacypreference",
        "providedidentity",
        ["fides_user_device_provided_identity_id"],
        ["id"],
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column(
            "fides_user_device",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("fides_user_device_provided_identity_id", sa.String(), nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("hashed_fides_user_device", sa.String(), nullable=True),
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_fides_user_device_provided_identity_id"),
        "privacypreferencehistory",
        ["fides_user_device_provided_identity_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_hashed_fides_user_device"),
        "privacypreferencehistory",
        ["hashed_fides_user_device"],
        unique=False,
    )
    op.create_foreign_key(
        "privacypreferencehistory_fides_user_device_provided_identi_fkey",
        "privacypreferencehistory",
        "providedidentity",
        ["fides_user_device_provided_identity_id"],
        ["id"],
    )


def downgrade():
    op.drop_constraint(
        "privacypreferencehistory_fides_user_device_provided_identi_fkey",
        "privacypreferencehistory",
        type_="foreignkey",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_hashed_fides_user_device"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_fides_user_device_provided_identity_id"),
        table_name="privacypreferencehistory",
    )
    op.drop_column("privacypreferencehistory", "hashed_fides_user_device")
    op.drop_column("privacypreferencehistory", "fides_user_device_provided_identity_id")
    op.drop_column("privacypreferencehistory", "fides_user_device")
    op.drop_constraint(
        "currentprivacypreference_fides_user_device_provided_identi_fkey",
        "currentprivacypreference",
        type_="foreignkey",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_fides_user_device_provided_identity_id"),
        table_name="currentprivacypreference",
    )
    op.drop_constraint(
        "fides_user_device_identity_privacy_notice",
        "currentprivacypreference",
        type_="unique",
    )
    op.drop_column("currentprivacypreference", "fides_user_device_provided_identity_id")



"""add website connection type

Revision ID: c961528edfc6
Revises: 021166731846
Create Date: 2025-01-27 19:20:20.236548

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c961528edfc6"
down_revision = "021166731846"
branch_labels = None
depends_on = None


def upgrade():
    # add 'website' to ConnectionType enum
    op.execute("ALTER TYPE connectiontype RENAME TO connectiontype_old")
    op.execute(
        """
        CREATE TYPE connectiontype AS ENUM (
            'mongodb',
            'mysql',
            'https',
            'snowflake',
            'redshift',
            'mssql',
            'mariadb',
            'bigquery',
            'saas',
            'manual',
            'manual_webhook',
            'timescale',
            'fides',
            'sovrn',
            'attentive_email',
            'dynamodb',
            'postgres',
            'generic_consent_email',
            'generic_erasure_email',
            'scylla',
            's3',
            'google_cloud_sql_mysql',
            'google_cloud_sql_postgres',
            'dynamic_erasure_email',
            'rds_mysql',
            'rds_postgres',
            'datahub',
            'website'
        )
    """
    )
    op.execute(
        """
        ALTER TABLE connectionconfig ALTER COLUMN connection_type TYPE connectiontype USING
        connection_type::text::connectiontype
    """
    )
    op.execute("DROP TYPE connectiontype_old")


def downgrade():
    # Remove 'website' from ConnectionType enum
    op.execute("DELETE FROM connectionconfig WHERE connection_type IN ('datahub')")
    op.execute("ALTER TYPE connectiontype RENAME TO connectiontype_old")
    op.execute(
        """
        CREATE TYPE connectiontype AS ENUM (
            'mongodb',
            'mysql',
            'https',
            'snowflake',
            'redshift',
            'mssql',
            'mariadb',
            'bigquery',
            'saas',
            'manual',
            'manual_webhook',
            'timescale',
            'fides',
            'sovrn',
            'attentive_email',
            'dynamodb',
            'postgres',
            'generic_consent_email',
            'generic_erasure_email',
            'scylla',
            's3',
            'google_cloud_sql_mysql',
            'google_cloud_sql_postgres',
            'dynamic_erasure_email',
            'rds_mysql',
            'rds_postgres',
            'datahub'
        )
    """
    )
    op.execute(
        """
        ALTER TABLE connectionconfig ALTER COLUMN connection_type TYPE connectiontype USING
        connection_type::text::connectiontype
    """
    )
    op.execute("DROP TYPE connectiontype_old")



"""This module handles configuring logging for the appliation."""

# pylint: disable=C0115,C0116, E0213
import os
from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING, getLevelName

from pydantic import Field, field_validator
from pydantic_settings import SettingsConfigDict

from .fides_settings import FidesSettings
from .utils import get_dev_mode

ENV_PREFIX = "FIDES__LOGGING__"


class LoggingSettings(FidesSettings):
    """Configuration settings for application logging."""

    # Logging
    destination: str = Field(
        default="",
        description="The output location for log files. Accepts any valid file path. If left unset, log entries are printed to stdout and log files are not produced.",
    )
    colorize: bool = Field(
        default=False,
        description="Force colored logs. Any value set via environment variables is considered 'True'.",
    )
    level: str = Field(
        default="INFO",
        description="The minimum log entry level to produce. Also accepts TRACE, DEBUG, WARNING, ERROR, or CRITICAL (case insensitive).",
    )
    serialization: str = Field(
        default="",
        description="The format with which to produce log entries. If left unset, produces log entries formatted using the internal custom formatter. Also accepts 'JSON' (case insensitive).",
    )
    log_pii: bool = Field(
        default=False,
        description="If True, PII values will display unmasked in log output. This variable should always be set to 'False' in production systems.",
    )

    @field_validator("destination", mode="before")
    @classmethod
    def get_destination(cls, value: str) -> str:
        """
        Print logs to sys.stdout, unless a valid file path is specified.
        """
        return value if os.path.exists(value) else ""

    @field_validator("level", mode="before")
    @classmethod
    def validate_log_level(cls, value: str) -> str:
        """Ensure the provided LEVEL is a valid value."""

        if get_dev_mode():
            return getLevelName(DEBUG)

        valid_values = [
            DEBUG,
            INFO,
            WARNING,
            ERROR,
            CRITICAL,
        ]
        value = value.upper()  # force uppercase for safety

        if getLevelName(value) not in valid_values:
            raise ValueError(
                f"Invalid LOG_LEVEL provided '{value}', must be one of: {', '.join([getLevelName(level) for level in valid_values])}"
            )

        return value

    @field_validator("serialization", mode="before")
    @classmethod
    def get_serialization(cls, value: str) -> str:
        """
        Ensure that only JSON serialization, or no serialization, is used.
        """
        value = value.lower()
        return value if value == "json" else ""

    model_config = SettingsConfigDict(env_prefix=ENV_PREFIX)



"""This module handles the logic required for pushing manifest files to the server."""

from pprint import pprint
from typing import Dict, List, Tuple

from deepdiff import DeepDiff
from fideslang.models import FidesModel, Taxonomy

from fides.common.utils import echo_green, echo_red, handle_cli_response
from fides.core import api
from fides.core.api_helpers import get_server_resources


def sort_create_update(
    manifest_resource_list: List[FidesModel],
    server_resource_list: List[FidesModel],
    diff: bool = False,
) -> Tuple[List[FidesModel], List[FidesModel]]:
    """
    Check the contents of the resource lists and populate separate
    new lists for resource creation or updating.

    The `diff` flag will print out the differences between the
    server resources and the local resource files.
    """

    server_resource_dict = {
        server_resource.fides_key: server_resource
        for server_resource in server_resource_list
    }

    create_list, update_list = [], []
    for manifest_resource in manifest_resource_list:
        resource_key = manifest_resource.fides_key

        # Check if the resource's fides_key matches one from the server
        if resource_key in server_resource_dict.keys():
            server_resource = server_resource_dict[resource_key]

            if diff:
                resource_diff = DeepDiff(
                    manifest_resource.model_dump(mode="json"),
                    server_resource.model_dump(mode="json"),
                )
                if resource_diff:
                    print(
                        f"\nUpdated resource with fides_key: {manifest_resource.fides_key}"
                    )
                    pprint(resource_diff)

            update_list.append(manifest_resource)

        else:
            if diff:
                print(f"\nNew resource with fides_key: {manifest_resource.fides_key}")
                pprint(manifest_resource.model_dump(exclude_unset=True))
            create_list.append(manifest_resource)

    return create_list, update_list


def echo_results(action: str, resource_type: str, resource_count: int) -> None:
    """
    Echo out the results of the push.
    """
    echo_green(f"{action.upper()} {resource_count} {resource_type} resource(s).")


def get_orphan_datasets(taxonomy: Taxonomy) -> List:
    """
    Validate all datasets are referenced at least one time
    by comparing all datasets to a set of referenced datasets
    within a privacy declaration.

    Returns a list containing any orphan datasets.
    """
    referenced_datasets = set()
    datasets = set()

    referenced_datasets.update(
        [
            dataset_reference
            for resource in taxonomy.system or []
            for privacy_declaration in resource.privacy_declarations
            if privacy_declaration.dataset_references is not None
            for dataset_reference in privacy_declaration.dataset_references
        ]
    )

    datasets.update([resource.fides_key for resource in taxonomy.dataset or []])

    missing_datasets = list(datasets - referenced_datasets)

    return missing_datasets


def push(
    url: str,
    taxonomy: Taxonomy,
    headers: Dict[str, str],
    dry: bool = False,
    diff: bool = False,
) -> None:
    """
    Push the current manifest file state to the server.
    """

    missing_datasets = get_orphan_datasets(taxonomy)
    if len(missing_datasets) > 0:
        echo_red(
            "Orphan Dataset Warning: The following datasets are not found referenced on a System"
        )
        for dataset in missing_datasets:
            print(dataset)

    for resource_type in taxonomy.model_fields_set:
        print("-" * 10)
        print(f"Processing {resource_type} resource(s)...")
        resource_list = getattr(taxonomy, resource_type)

        if diff or dry:
            existing_keys = [resource.fides_key for resource in resource_list]
            server_resource_list = get_server_resources(
                url, resource_type, existing_keys, headers
            )

            # Determine which resources should be created, updated, or are unchanged
            create_list, update_list = sort_create_update(
                resource_list, server_resource_list, diff
            )

            if dry:
                echo_results("would create", resource_type, len(create_list))
                echo_results("would update", resource_type, len(update_list))
                continue

        handle_cli_response(
            api.upsert(
                headers=headers,
                resource_type=resource_type,
                url=url,
                resources=[
                    resource.model_dump(mode="json") for resource in resource_list
                ],
            ),
            verbose=False,
        )

        echo_results("pushed", resource_type, len(resource_list))

    print("-" * 10)



"""empty message

Revision ID: 49bdd2fff350
Revises: c9abc9bbaa49
Create Date: 2024-10-09 16:02:06.996661

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "49bdd2fff350"
down_revision = "c9abc9bbaa49"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        WITH adobe_campaign_connections AS (
            SELECT id
            FROM connectionconfig
            WHERE connection_type = 'saas'
              AND saas_config->>'type' = 'adobe_campaign'
        )
        DELETE FROM datasetconfig
        WHERE connection_config_id IN (SELECT id FROM adobe_campaign_connections);

        DELETE FROM connectionconfig WHERE connection_type = 'saas' AND saas_config->>'type' = 'adobe_campaign';
    """
    )


def downgrade():
    # cannot downgrade since we can't restore a deleted connectionconfig
    pass



from __future__ import annotations

from enum import Enum
from typing import Dict, List, Optional

from pydantic import ConfigDict, Field, SerializeAsAny, field_validator, model_validator

from fides.api.custom_types import AnyHttpUrlStringRemovesSlash, URLOriginString
from fides.api.schemas.base_class import FidesSchema
from fides.api.schemas.messaging.messaging import MessagingServiceType


class StorageTypeApiAccepted(Enum):
    """Enum for storage destination types accepted in API updates"""

    s3 = "s3"
    local = "local"  # local should be used for testing only, not for processing real-world privacy requests


class StorageApplicationConfig(FidesSchema):
    active_default_storage_type: StorageTypeApiAccepted
    model_config = ConfigDict(use_enum_values=True, extra="forbid")


# TODO: the below models classes are "duplicates" of the pydantic
# models that drive the application config. this is to allow every field
# to be optional on the API model, since we want PATCH functionality.
# ideally, we'd not need to duplicate the config modelclasses, and instead
# just make all fields optional by default for the API models.


class NotificationApplicationConfig(FidesSchema):
    """
    API model - configuration settings for data subject and/or data processor notifications
    """

    send_request_completion_notification: Optional[bool] = None
    send_request_receipt_notification: Optional[bool] = None
    send_request_review_notification: Optional[bool] = None
    notification_service_type: Optional[str] = None
    enable_property_specific_messaging: Optional[bool] = None
    model_config = ConfigDict(extra="forbid")

    @field_validator("notification_service_type", mode="before")
    @classmethod
    def validate_notification_service_type(cls, value: str) -> Optional[str]:
        """Ensure the provided type is a valid value."""
        value = value.lower()  # force lowercase for safety
        try:
            MessagingServiceType[value]
        except KeyError:
            raise ValueError(
                f"Invalid notification.notification_service_type provided '{value}', must be one of: {', '.join([service_type.name for service_type in MessagingServiceType])}"
            )

        return value


class ExecutionApplicationConfig(FidesSchema):
    subject_identity_verification_required: Optional[bool] = None
    disable_consent_identity_verification: Optional[bool] = None
    require_manual_request_approval: Optional[bool] = None
    model_config = ConfigDict(extra="forbid")


class AdminUIConfig(FidesSchema):
    enabled: Optional[bool] = None
    url: SerializeAsAny[Optional[AnyHttpUrlStringRemovesSlash]] = None

    model_config = ConfigDict(extra="forbid")


class ConsentConfig(FidesSchema):
    override_vendor_purposes: Optional[bool]
    model_config = ConfigDict(extra="forbid")


class SecurityApplicationConfig(FidesSchema):
    # only valid URLs should be set as cors_origins
    # for advanced usage of non-URLs, e.g. wildcards (`*`), the related
    # `cors_origin_regex` property should be used.
    # this is explicitly _not_ accessible via API - it must be used with care.
    cors_origins: SerializeAsAny[Optional[List[URLOriginString]]] = Field(
        default=None,
        description="A list of client addresses allowed to communicate with the Fides webserver.",
    )
    model_config = ConfigDict(extra="forbid")


class ApplicationConfig(FidesSchema):
    """
    Application config settings update body is an arbitrary dict (JSON object)
    We describe it in a schema to enforce some restrictions on the keys passed.

    TODO: Eventually this should be driven by a more formal validation schema for this
    the application config that is properly hooked up to the global pydantic config module.
    """

    storage: Optional[StorageApplicationConfig] = None
    notifications: Optional[NotificationApplicationConfig] = None
    execution: Optional[ExecutionApplicationConfig] = None
    security: Optional[SecurityApplicationConfig] = None
    consent: Optional[ConsentConfig] = None
    admin_ui: Optional[AdminUIConfig] = None

    @model_validator(mode="before")
    @classmethod
    def validate_not_empty(cls, values: Dict) -> Dict:
        if not values:
            raise ValueError(
                "Config body cannot be empty. DELETE endpoint can be used to null out application config."
            )
        return values

    model_config = ConfigDict(extra="forbid")



from sqlalchemy.orm import Session

from fides.api.util.fuzzy_search_utils import (
    add_identity_to_automaton,
    get_decrypted_identities_automaton,
)


class DecryptedIdentityAutomatonMixin:
    """
    A class housing common decrypted identity automaton logic
    """

    def add_identities_to_automaton(self) -> None:
        """
        Manually add identities to automaton. Currently, used only during privacy request creation.

        If the automaton has expired, this method also refreshes the entire automaton.
        """
        db = Session.object_session(self)
        automaton = get_decrypted_identities_automaton(db, True)
        add_identity_to_automaton(automaton, self.id, self.get_persisted_identity().__dict__)  # type: ignore



import json
from typing import Dict, List, Optional

from requests import PreparedRequest

from fides.api.common_exceptions import FidesopsException
from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.schemas.saas.saas_config import Header, QueryParam
from fides.api.schemas.saas.strategy_configuration import (
    ApiKeyAuthenticationConfiguration,
)
from fides.api.service.authentication.authentication_strategy import (
    AuthenticationStrategy,
)
from fides.api.util.saas_util import assign_placeholders
from fides.api.util.url_util import set_query_parameter


class ApiKeyAuthenticationStrategy(AuthenticationStrategy):
    """
    Adds an API key to each request header, query param, or the body
    depending on configuration
    """

    name = "api_key"
    configuration_model = ApiKeyAuthenticationConfiguration

    def __init__(self, configuration: ApiKeyAuthenticationConfiguration):
        self.headers: List[Header] = configuration.headers  # type: ignore
        self.query_params: List[QueryParam] = configuration.query_params  # type: ignore
        self.body: Optional[str] = configuration.body  # type: ignore

    def add_authentication(
        self, request: PreparedRequest, connection_config: ConnectionConfig
    ) -> PreparedRequest:
        """
        Add API key authentication to the request header or query param
        """
        secrets = connection_config.secrets
        if not secrets:
            raise FidesopsException(
                "Secrets are not configured for this SaaS connector. Secrets must be configured to use API key authentication"
            )
        if self.headers:
            for header in self.headers:
                header_val = assign_placeholders(header.value, secrets)
                if header_val is None:
                    raise FidesopsException(
                        f"Value for API key param '{header.value}' not found"
                    )
                request.headers[header.name] = header_val
        if self.query_params:
            for query_param in self.query_params:
                param_val = assign_placeholders(query_param.value, secrets)
                if param_val is None:
                    raise FidesopsException(
                        f"Value for API key param '{query_param.value}' not found"
                    )
                request.url = set_query_parameter(
                    request.url,  # type: ignore
                    query_param.name,
                    param_val,
                )
        if self.body:
            body_with_api_key: Optional[str] = assign_placeholders(
                self.body, secrets or {}
            )
            api_key_request_body: Dict = json.loads(body_with_api_key or "{}")

            original_request_body: Dict = json.loads(request.body or "{}")
            original_request_body.update(api_key_request_body)

            request.prepare_body(
                data=json.dumps(original_request_body), files=None, json=True
            )

        return request



"""update messaging service_type enum

Revision ID: 2f6aa5fe797a
Revises: d65bbc647083
Create Date: 2023-03-03 17:13:21.108111

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "2f6aa5fe797a"
down_revision = "39b209861471"
branch_labels = None
depends_on = None


def upgrade():
    # create a temp enum type for old values
    op.execute("alter type messagingservicetype rename to messagingservicetype_old")
    # create a temp column for old values
    op.alter_column(
        "messagingconfig", "service_type", new_column_name="service_type_old"
    )
    # update the temp column to use the temp enum type
    op.execute(
        (
            "alter table messagingconfig alter column service_type_old type messagingservicetype_old using "
            "service_type_old::text::messagingservicetype_old"
        )
    )

    # create the new enum type
    messagingservicetype = postgresql.ENUM(
        "mailgun",
        "twilio_text",
        "twilio_email",
        "mailchimp_transactional",
        name="messagingservicetype",
        create_type=True,
    )
    messagingservicetype.create(op.get_bind())

    # create the new column using the new enum type
    op.add_column(
        "messagingconfig",
        sa.Column(
            "service_type",
            messagingservicetype,
            nullable=True,  # allow it to be nullable temporarily
            unique=True,
        ),
    )

    # now actually migrate the old values to the new column
    session = Session(bind=op.get_bind())
    conn = session.connection()
    statement = text(
        """SELECT distinct LOWER(service_type_old::text) AS lower_service_type_old, id
FROM messagingconfig"""
    )

    result = conn.execute(statement)
    for row in result:
        op.execute(
            f"UPDATE messagingconfig SET service_type = '{row[0]}' WHERE id = '{row[1]}';"
        )

    # update our column to make it non-nullable now that it's populated
    op.alter_column("messagingconfig", "service_type", nullable=False)

    # and clean up our temporary resources
    op.drop_column("messagingconfig", "service_type_old")
    op.execute("drop type messagingservicetype_old")


def downgrade():
    # create a temp enum type for old values
    op.execute("alter type messagingservicetype rename to messagingservicetype_old")
    # create a temp column for old values
    op.alter_column(
        "messagingconfig", "service_type", new_column_name="service_type_old"
    )
    # update the temp column to use the temp enum type
    op.execute(
        (
            "alter table messagingconfig alter column service_type_old type messagingservicetype_old using "
            "service_type_old::text::messagingservicetype_old"
        )
    )

    # create the new column with the new enum type
    messagingservicetype = postgresql.ENUM(
        "MAILGUN",
        "TWILIO_TEXT",
        "TWILIO_EMAIL",
        "MAILCHIMP_TRANSACTIONAL",
        name="messagingservicetype",
        create_type=True,
    )
    messagingservicetype.create(op.get_bind())
    op.add_column(
        "messagingconfig",
        sa.Column(
            "service_type",
            messagingservicetype,
            nullable=True,  # allow it to be nullable temporarily
            unique=True,
        ),
    )

    # now actually migrate the old values to the new column
    session = Session(bind=op.get_bind())
    conn = session.connection()
    statement = text(
        """SELECT distinct UPPER(service_type_old::text) AS upper_service_type_old, id
FROM messagingconfig"""
    )

    result = conn.execute(statement)
    for row in result:
        op.execute(
            f"UPDATE messagingconfig SET service_type = '{row[0]}' WHERE id = '{row[1]}';"
        )

    # update our column to make it non-nullable now that it's populated
    op.alter_column("messagingconfig", "service_type", nullable=False)

    # and clean up our temporary resources
    op.drop_column("messagingconfig", "service_type_old")
    op.execute("drop type messagingservicetype_old")



import json
from typing import ClassVar, List, Optional, Union

from pydantic import EmailStr, Field, field_validator
from pydantic.main import BaseModel

from fides.api.schemas.base_class import NoValidationSchema
from fides.api.schemas.connection_configuration.connection_secrets import (
    ConnectionConfigSecretsSchema,
)


class KeyfileCreds(BaseModel):
    """Schema that holds Google Cloud SQL for MySQL keyfile key/vals"""

    type: Optional[str] = None
    project_id: str = Field(title="Project ID")
    private_key_id: Optional[str] = Field(default=None, title="Private key ID")
    private_key: Optional[str] = Field(
        default=None, json_schema_extra={"sensitive": True}
    )
    client_email: Optional[EmailStr] = None
    client_id: Optional[str] = Field(default=None, title="Client ID")
    auth_uri: Optional[str] = Field(default=None, title="Auth URI")
    token_uri: Optional[str] = Field(default=None, title="Token URI")
    auth_provider_x509_cert_url: Optional[str] = Field(
        None, title="Auth provider X509 cert URL"
    )
    client_x509_cert_url: Optional[str] = Field(
        default=None, title="Client X509 cert URL"
    )
    universe_domain: str = Field(title="Universe domain")


class GoogleCloudSQLMySQLSchema(ConnectionConfigSecretsSchema):
    """Schema to validate the secrets needed to connect to Google Cloud SQL MySQL"""

    db_iam_user: str = Field(
        title="DB IAM user",
        description="example: service-account@test.iam.gserviceaccount.com",
    )
    instance_connection_name: str = Field(
        title="Instance connection name",
        description="example: friendly-tower-424214-n8:us-central1:test-ethyca",
    )
    dbname: str = Field(
        title="Database name",
    )
    keyfile_creds: KeyfileCreds = Field(
        title="Keyfile creds",
        json_schema_extra={"sensitive": True},
        description="The contents of the key file that contains authentication credentials for a service account in GCP.",
    )

    _required_components: ClassVar[List[str]] = [
        "db_iam_user",
        "instance_connection_name",
        "dbname",
        "keyfile_creds",
    ]

    @field_validator("keyfile_creds", mode="before")
    @classmethod
    def parse_keyfile_creds(cls, v: Union[str, dict]) -> KeyfileCreds:
        if isinstance(v, str):
            v = json.loads(v)
        return KeyfileCreds.model_validate(v)


class GoogleCloudSQLMySQLDocsSchema(GoogleCloudSQLMySQLSchema, NoValidationSchema):
    """Google Cloud SQL MySQL Secrets Schema for API Docs"""



"""fix monitorexecution.started default

Revision ID: 1152c1717849
Revises: 3c58001ad310
Create Date: 2025-03-10 16:04:57.104689

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "1152c1717849"
down_revision = "3c58001ad310"
branch_labels = None
depends_on = None


def upgrade():
    # Set the new server default
    # And set the datetime to be timezone aware
    op.alter_column(
        "monitorexecution",
        "started",
        existing_type=sa.DateTime(),
        type_=sa.DateTime(timezone=True),
        server_default=sa.text("now()"),
    )

    # Set the datetime to be timezone aware
    op.alter_column(
        "monitorexecution",
        "completed",
        existing_type=sa.DateTime(),
        type_=sa.DateTime(timezone=True),
    )


def downgrade():
    # Remove the server default
    # And set the datetime to not be timezone aware
    op.alter_column(
        "monitorexecution",
        "started",
        server_default=None,
        existing_type=sa.DateTime(timezone=True),
        type_=sa.DateTime(),
    )

    # Set the datetime to not be timezone aware
    op.alter_column(
        "monitorexecution",
        "completed",
        existing_type=sa.DateTime(timezone=True),
        type_=sa.DateTime(),
    )



from typing import ClassVar, List, Optional

from pydantic import Field, model_validator

from fides.api.schemas.base_class import NoValidationSchema
from fides.api.schemas.connection_configuration.connection_secrets import (
    ConnectionConfigSecretsSchema,
)


def format_private_key(raw_key: str) -> str:
    # Split the key into parts and remove spaces from the key body
    parts = raw_key.split("-----")
    body = parts[2].replace(" ", "\n")
    # Reassemble the key
    return f"-----{parts[1]}-----{body}-----{parts[3]}-----"


class SnowflakeSchema(ConnectionConfigSecretsSchema):
    """Schema to validate the secrets needed to connect to Snowflake"""

    account_identifier: str = Field(
        title="Account Name",
        description="The unique identifier for your Snowflake account.",
    )
    user_login_name: str = Field(
        title="Username",
        description="The user account used to authenticate and access the database.",
    )
    password: Optional[str] = Field(
        title="Password",
        description="The password used to authenticate and access the database. You can use a password or a private key, but not both.",
        default=None,
        json_schema_extra={"sensitive": True},
    )
    private_key: Optional[str] = Field(
        title="Private key",
        description="The private key used to authenticate and access the database. If a `private_key_passphrase` is also provided, it is assumed to be encrypted; otherwise, it is assumed to be unencrypted.",
        default=None,
        json_schema_extra={"sensitive": True},
    )
    private_key_passphrase: Optional[str] = Field(
        title="Passphrase",
        description="The passphrase used for the encrypted private key.",
        default=None,
        json_schema_extra={"sensitive": True},
    )
    warehouse_name: str = Field(
        title="Warehouse",
        description="The name of the Snowflake warehouse where your queries will be executed.",
    )
    database_name: Optional[str] = Field(
        default=None,
        title="Database",
        description="Only provide a database name to scope discovery monitors and privacy request automation to a specific database. In most cases, this can be left blank.",
    )
    schema_name: Optional[str] = Field(
        default=None,
        title="Schema",
        description="Only provide a schema to scope discovery monitors and privacy request automation to a specific schema. In most cases, this can be left blank.",
    )
    role_name: Optional[str] = Field(
        title="Role",
        default=None,
        description="The Snowflake role to assume for the session, if different than Username.",
    )

    _required_components: ClassVar[List[str]] = [
        "account_identifier",
        "user_login_name",
        "warehouse_name",
    ]

    @model_validator(mode="after")
    def validate_private_key_and_password(self) -> "SnowflakeSchema":
        private_key: str = self.private_key or ""

        if self.password and private_key:
            raise ValueError(
                "Cannot provide both password and private key at the same time."
            )

        if not any([self.password, private_key]):
            raise ValueError("Must provide either a password or a private key.")

        if private_key:
            try:
                self.private_key = format_private_key(private_key)
            except IndexError:
                raise ValueError("Invalid private key format")

        return self


class SnowflakeDocsSchema(SnowflakeSchema, NoValidationSchema):
    """Snowflake Secrets Schema for API Docs"""



"""add modal_link_label to experience translation model

Revision ID: e4023342ebbb
Revises: 7799269e7078
Create Date: 2024-04-02 01:16:46.183952

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "e4023342ebbb"
down_revision = "7799269e7078"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "experiencetranslation",
        sa.Column("modal_link_label", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("modal_link_label", sa.String(), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("privacyexperienceconfighistory", "modal_link_label")
    op.drop_column("experiencetranslation", "modal_link_label")
    # ### end Alembic commands ###



from typing import List, Optional, Type

from fideslang.models import Dataset as FideslangDataset
from fideslang.models import DatasetCollection, DatasetField
from fideslang.validation import FidesKey
from loguru import logger
from pydantic import BaseModel, field_validator

from fides.api.common_exceptions import DataCategoryNotSupported
from fides.api.util.data_category import DataCategory as DefaultTaxonomyDataCategories
from fides.api.util.data_category import get_data_categories_from_db
from fides.service.dataset.dataset_validator import (
    DatasetValidationContext,
    DatasetValidationStep,
)


def _valid_data_categories(
    proposed_data_categories: Optional[List[FidesKey]],
    defined_data_categories: List[FidesKey],
) -> Optional[List[FidesKey]]:
    """
    Ensure that every data category provided matches a valid defined data category.
    Throws an error if any of the categories are invalid,
    or otherwise returns the list of categories unchanged.
    """

    def validate_category(data_category: FidesKey) -> FidesKey:
        if data_category not in defined_data_categories:
            raise DataCategoryNotSupported(
                f"The data category {data_category} is not supported."
            )
        return data_category

    categories = proposed_data_categories or []
    for category in categories:
        validate_category(category)

    return categories


def validate_data_categories_against_db(
    dataset: FideslangDataset, defined_data_categories: List[FidesKey]
) -> None:
    """
    Validate that data_categories defined on the Dataset, Collection, and Field levels exist
    in the database.  Doing this instead of a traditional validator function to have
    access to a database session.

    If no data categories in the database, default to using data categories from the default taxonomy.
    """
    if not defined_data_categories:
        logger.info(
            "No data categories in the database: reverting to default data categories."
        )
        defined_data_categories = [
            FidesKey(key) for key in DefaultTaxonomyDataCategories.__members__.keys()
        ]

    class DataCategoryValidationMixin(BaseModel):
        @field_validator("data_categories", check_fields=False)
        @classmethod
        def valid_data_categories(
            cls: Type["DataCategoryValidationMixin"], v: Optional[List[FidesKey]]
        ) -> Optional[List[FidesKey]]:
            """Validate that all annotated data categories exist in the taxonomy"""
            return _valid_data_categories(v, defined_data_categories)

    class FieldDataCategoryValidation(DatasetField, DataCategoryValidationMixin):
        fields: Optional[List["FieldDataCategoryValidation"]] = None  # type: ignore[assignment]

    FieldDataCategoryValidation.model_rebuild()

    class CollectionDataCategoryValidation(
        DatasetCollection, DataCategoryValidationMixin
    ):
        fields: List[FieldDataCategoryValidation] = []  # type: ignore[assignment]

    class DatasetDataCategoryValidation(FideslangDataset, DataCategoryValidationMixin):
        collections: List[CollectionDataCategoryValidation]  # type: ignore[assignment]

    DatasetDataCategoryValidation(**dataset.model_dump(mode="json"))


class DataCategoryValidationStep(DatasetValidationStep):
    """Validates data categories against database"""

    def validate(self, context: DatasetValidationContext) -> None:
        defined_data_categories = get_data_categories_from_db(context.db)
        validate_data_categories_against_db(context.dataset, defined_data_categories)



from typing import Any, Dict, List, Optional, Union

import pydash
from loguru import logger

from fides.api.common_exceptions import FidesopsException
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.saas.shared_schemas import DatasetRef, IdentityParamRef
from fides.api.schemas.saas.strategy_configuration import (
    FilterPostProcessorConfiguration,
)
from fides.api.service.processors.post_processor_strategy.post_processor_strategy import (
    PostProcessorStrategy,
)


class FilterPostProcessorStrategy(PostProcessorStrategy):
    """
    Filters object or array given field name and value
    Value can be reference a dynamic identity passed in through the request OR hard-coded value.
    E.g.
    data = [
        {
            "id": 1397429347
            "email_contact": somebody@email.com
            "name": Somebody Awesome
        },
        {
            "id": 238475234
            "email_contact": somebody-else@email.com
            "name": Somebody Cool
        }
    ]
    field: email_contact
    value: {"identity": email}, where email == somebody@email.com
    result = {
        id: 1397429347
        email_contact: somebody@email.com
        name: Somebody Awesome
    }
    """

    name = "filter"
    configuration_model = FilterPostProcessorConfiguration

    def __init__(self, configuration: FilterPostProcessorConfiguration):
        self.field = configuration.field
        self.value = configuration.value
        self.exact = configuration.exact
        self.case_sensitive = configuration.case_sensitive

    def process(
        self,
        data: Union[List[Dict[str, Any]], Dict[str, Any]],
        identity_data: Optional[Dict[str, Any]] = None,
        privacy_request: Optional[PrivacyRequest] = None,
    ) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        """
        - data: A list or a dict
        - identity_data: A dict of cached identity data
        - privacy_request: A Privacy Request object

        returns a filtered dict, list of dicts, or empty list
        """

        if not data:
            return []

        # find value to filter by
        filter_value = self.value
        if isinstance(self.value, IdentityParamRef):
            if identity_data is None or identity_data.get(self.value.identity) is None:
                logger.warning(
                    "Could not retrieve identity reference '{}' due to missing identity data for the following post processing strategy: {}",
                    self.value.identity,
                    self.name,
                )
                return []
            filter_value = identity_data.get(self.value.identity)  # type: ignore

        if isinstance(self.value, DatasetRef):
            if (
                privacy_request is None
                or len(self.value.dataset_reference.split(".")) != 3
            ):
                logger.warning(
                    "Could not retrieve dataset reference '{}' due to missing collection data or wrong dataset format for the following post processing strategy: {}",
                    self.value.dataset_reference,
                    self.name,
                )
                return []
            access_data = privacy_request.get_raw_access_results()
            dataset_reference = self.value.dataset_reference.split(".")
            dataset, collection, field = dataset_reference
            filter_value = self._get_nested_values(
                access_data, f"{dataset}:{collection}.{field}"
            )

        try:
            if isinstance(data, list):
                return [
                    item
                    for item in data
                    if self._matches(
                        self.exact,
                        self.case_sensitive,
                        filter_value if isinstance(filter_value, list) else [filter_value],  # type: ignore
                        self._get_nested_values(item, self.field),
                    )
                ]
            return (
                data
                if self._matches(
                    self.exact,
                    self.case_sensitive,
                    filter_value if isinstance(filter_value, list) else [filter_value],  # type: ignore
                    self._get_nested_values(data, self.field),
                )
                else []
            )
        except KeyError:
            logger.warning(
                "{} could not be found on data for the following post processing strategy: {}",
                self.field,
                self.name,
            )
            return []

    def _matches(
        self,
        exact: bool,
        case_sensitive: bool,
        filter_value: Union[List[str], List[int]],
        target: Union[str, List[str], int, List[int]],
    ) -> bool:
        """
        Returns a boolean indicating if the filter_value (list[string] or list[int]) is contained
        in the target (string, list of strings, int, list of ints).

        - exact: filter_value and target must be the same length (no extra characters)
        - case_sensitive: cases must match between filter_value and target
        """

        # does not match if we don't have anything to compare to
        if target is None:
            return False

        # validate inputs
        if not isinstance(target, (str, list, int)):
            raise FidesopsException(
                f"Field value '{self.field}' for filter postprocessor must be a string, list of strings, integer or list of integers, found '{type(target).__name__}'"
            )

        # validate list contents
        if isinstance(target, list):
            if not all(isinstance(item, str) for item in target) and not all(
                isinstance(item, int) for item in target
            ):
                raise FidesopsException(
                    f"The field '{self.field}' list must contain either all strings or all integers."
                )
        # validate filter contents
        if not all(isinstance(value, str) for value in filter_value) and not all(
            isinstance(value, int) for value in filter_value
        ):
            raise FidesopsException(
                f"The filter_value '{filter_value}' list must contain either all strings or all integers."
            )

        if isinstance(target, int):
            return any(value == target for value in filter_value)

        if isinstance(target, list) and isinstance(target[0], int):
            return any(value in target for value in filter_value)

        # prep inputs by converting them to lowercase
        if not case_sensitive and isinstance(target, (list, str)):
            filter_value = [
                value.casefold() for value in filter_value if isinstance(value, str)
            ]
            if isinstance(target, list) and isinstance(target[0], str):
                target = [item.casefold() for item in target if isinstance(item, str)]
            elif isinstance(target, str):
                target = target.casefold()

        # compare filter_values to a target list
        if isinstance(target, list) and isinstance(target[0], str):
            return any(
                value == item if exact else value in item
                for value in filter_value
                if isinstance(value, str)
                for item in target
                if isinstance(item, str)
            )

        # base case, compare filter_value to a single string
        return any(
            value == target if exact else value in target for value in filter_value
        )

    def _get_nested_values(self, data: Dict[str, Any], path: str) -> Any:
        """
        Extracts nested values from a dictionary based on a dot-separated path string.

        This function handles cases where the path leads to a list of dictionaries.
        Instead of returning the list, it iterates through each dictionary in the list,
        applies the remaining path, and returns a flattened list of the results.

        Parameters:
        data (dict): The dictionary from which to extract the value.
        path (str): The dot-separated string path indicating the nested value.

        Returns:
        The value(s) at the nested path within the dictionary. If the path
        leads to a list of dictionaries, it returns a flattened list of values.

        Raises:
        KeyError: If the path does not exist in the dictionary.

        Example:
        For a dictionary {"a": {"b": [{"c": 1}, {"c": 2}]}} and path "a.b.c", the function
        returns [1, 2].
        """
        components = path.split(".")
        value = data[components[0]]
        if len(components) > 1:
            if isinstance(value, dict):
                return self._get_nested_values(value, ".".join(components[1:]))
            if isinstance(value, list):
                return pydash.flatten(
                    [
                        self._get_nested_values(item, ".".join(components[1:]))
                        for item in value
                    ]
                )
        return value



"""Add sovrn consent connector

Revision ID: 8e198eb13802
Revises: 643249f65453
Create Date: 2023-01-27 21:58:23.344582

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "8e198eb13802"
down_revision = "643249f65453"
branch_labels = None
depends_on = None
import sqlalchemy as sa


def upgrade():
    # Add to ConnectionType enum
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', 'snowflake', 'redshift', 'mssql', 'mariadb', 'bigquery', 'saas', 'manual', 'email', 'manual_webhook', 'timescale', 'fides', 'sovrn')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")

    # Add new PrivacyRequest.awaiting_consent_email_send_at column
    op.add_column(
        "privacyrequest",
        sa.Column(
            "awaiting_consent_email_send_at", sa.DateTime(timezone=True), nullable=True
        ),
    )

    # Add ljt_reader_id for sovrn
    op.execute("ALTER TYPE providedidentitytype ADD VALUE 'ljt_readerID'")

    # Add new privacyrequeststatus
    op.execute(
        "alter type privacyrequeststatus add value 'awaiting_consent_email_send'"
    )


def downgrade():
    # Remove sovrn from the connectiontype enum
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', 'snowflake', 'redshift', 'mssql', 'mariadb', 'bigquery', 'saas', 'manual', 'email', 'manual_webhook', 'timescale', 'fides')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")

    # # Drop PrivacyRequest.awaiting_consent_email_send_at column
    op.drop_column("privacyrequest", "awaiting_consent_email_send_at")

    # Remove ljt_reader_id from the providedidentitytype enum
    op.execute("ALTER TYPE providedidentitytype RENAME TO providedidentitytype_old")
    op.execute(
        "CREATE TYPE providedidentitytype AS ENUM('email', 'phone_number', 'ga_client_id')"
    )
    op.execute(
        "ALTER TABLE providedidentity ALTER COLUMN field_name TYPE providedidentitytype USING field_name::text::providedidentitytype"
    )
    op.execute("DROP TYPE providedidentitytype_old")

    # Removing awaiting_consent_email_send privacyrequeststatus
    op.execute(
        "delete from privacyrequest where status in ('awaiting_consent_email_send')"
    )
    op.execute("alter type privacyrequeststatus rename to privacyrequeststatus_old")
    op.execute(
        "create type privacyrequeststatus as enum('in_processing', 'complete', 'pending', 'error', 'paused', 'approved', 'denied', 'canceled', 'identity_unverified', 'requires_input')"
    )
    op.execute(
        (
            "alter table privacyrequest alter column status type privacyrequeststatus using "
            "status::text::privacyrequeststatus"
        )
    )
    op.execute("drop type privacyrequeststatus_old")



# pylint: disable=too-many-lines
from typing import Any, Dict, List, Optional, TypeVar

from sqlalchemy.sql import Executable  # type: ignore

from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.privacy_request import ManualAction
from fides.api.service.connectors.query_configs.query_config import QueryConfig
from fides.api.util.collection_util import Row, filter_nonempty_values

T = TypeVar("T")


class ManualQueryConfig(QueryConfig[Executable]):
    def generate_query(
        self, input_data: Dict[str, List[Any]], policy: Optional[Policy]
    ) -> Optional[ManualAction]:
        """Describe the details needed to manually retrieve data from the
        current collection.

        Example:
        {
            "step": "access",
            "collection": "manual_dataset:manual_collection",
            "action_needed": [
                {
                    "locators": {'email': "customer-1@example.com"},
                    "get": ["id", "box_id"]
                    "update":  {}
                }
            ]
        }

        """

        locators: Dict[str, Any] = self.node.typed_filtered_values(input_data)
        get: List[str] = [
            field_path.string_path
            for field_path in self.node.collection.top_level_field_dict
        ]

        if get and locators:
            return ManualAction(locators=locators, get=get, update=None)
        return None

    def query_to_str(self, t: T, input_data: Dict[str, List[Any]]) -> None:
        """Not used for ManualQueryConfig, we output the dry run query as a dictionary instead of a string"""

    def dry_run_query(self) -> Optional[ManualAction]:  # type: ignore
        """Displays the ManualAction needed with question marks instead of action data for the locators
        as a dry run query"""
        fake_data: Dict[str, Any] = self.display_query_data()
        manual_query: Optional[ManualAction] = self.generate_query(fake_data, None)
        if not manual_query:
            return None

        for where_params in manual_query.locators.values():
            for i, _ in enumerate(where_params):
                where_params[i] = "?"
        return manual_query

    def generate_update_stmt(
        self, row: Row, policy: Policy, request: PrivacyRequest
    ) -> Optional[ManualAction]:
        """Describe the details needed to manually mask data in the
        current collection.

        Example:
         {
            "step": "erasure",
            "collection": "manual_dataset:manual_collection",
            "action_needed": [
                {
                    "locators": {'id': 1},
                    "get": []
                    "update":  {'authorized_user': None}
                }
            ]
        }
        """
        locators: Dict[str, Any] = filter_nonempty_values(
            {
                field_path.string_path: field.cast(row[field_path.string_path])
                for field_path, field in self.primary_key_field_paths.items()
            }
        )
        update_stmt: Dict[str, Any] = self.update_value_map(row, policy, request)

        if update_stmt and locators:
            return ManualAction(locators=locators, get=None, update=update_stmt)
        return None



"""add cookie table

Revision ID: 2be84e68df32
Revises: c1885270b3cc
Create Date: 2023-06-13 23:08:35.011377

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "2be84e68df32"
down_revision = "c1885270b3cc"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "cookies",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("domain", sa.String(), nullable=True),
        sa.Column("path", sa.String(), nullable=True),
        sa.Column("system_id", sa.String(), nullable=True),
        sa.Column("privacy_declaration_id", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["privacy_declaration_id"], ["privacydeclaration.id"], ondelete="SET NULL"
        ),
        sa.ForeignKeyConstraint(["system_id"], ["ctl_systems.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "name", "privacy_declaration_id", name="_cookie_name_privacy_declaration_uc"
        ),
    )
    op.create_index(op.f("ix_cookies_id"), "cookies", ["id"], unique=False)
    op.create_index(op.f("ix_cookies_name"), "cookies", ["name"], unique=False)
    op.create_index(
        op.f("ix_cookies_privacy_declaration_id"),
        "cookies",
        ["privacy_declaration_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_cookies_system_id"), "cookies", ["system_id"], unique=False
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_cookies_system_id"), table_name="cookies")
    op.drop_index(op.f("ix_cookies_privacy_declaration_id"), table_name="cookies")
    op.drop_index(op.f("ix_cookies_name"), table_name="cookies")
    op.drop_index(op.f("ix_cookies_id"), table_name="cookies")
    op.drop_table("cookies")
    # ### end Alembic commands ###



"""Add last_monitored and enabled to MonitorConfig

Revision ID: 4fb779cc5f17
Revises: 8976def679db
Create Date: 2024-06-21 13:45:46.024496

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4fb779cc5f17"
down_revision = "8976def679db"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "monitorconfig",
        sa.Column("last_monitored", sa.DateTime(timezone=True), nullable=True),
    )
    op.add_column(
        "monitorconfig",
        sa.Column("enabled", sa.Boolean(), nullable=False, server_default="t"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("monitorconfig", "enabled")
    op.drop_column("monitorconfig", "last_monitored")
    # ### end Alembic commands ###



"""
emailconfig

Revision ID: c61fd9d4f73e
Revises: 7abe778b7082
Create Date: 2022-08-10 19:29:12.119401

"""

from alembic import op

# revision identifiers, used by Alembic.

revision = "c61fd9d4f73e"
down_revision = "7abe778b7082"
branch_labels = None
depends_on = None


def upgrade():
    # Removing due to the following error: type "emailservicetype" already exists
    pass


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_emailconfig_service_type"), table_name="emailconfig")
    op.drop_index(op.f("ix_emailconfig_name"), table_name="emailconfig")
    op.drop_index(op.f("ix_emailconfig_key"), table_name="emailconfig")
    op.drop_index(op.f("ix_emailconfig_id"), table_name="emailconfig")
    op.drop_table("emailconfig")
    # ### end Alembic commands ###



from typing import List

from pydantic import BaseModel


class MaskingStrategyConfigurationDescription(BaseModel):
    """The description model for a specific configuration in a masking strategy"""

    key: str
    optional: bool = True
    description: str


class MaskingStrategyDescription(BaseModel):
    """The description model for a masking strategy"""

    name: str
    description: str
    configurations: List[MaskingStrategyConfigurationDescription]



"""add plus_privacy_experience_config table

Revision ID: 69e51a460e66
Revises: 3815b0db5b14
Create Date: 2024-03-01 22:47:59.867223

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "69e51a460e66"
down_revision = "3815b0db5b14"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "plus_privacy_experience_config_property",
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("id", sa.String(length=255), nullable=True),
        sa.Column("privacy_experience_config_id", sa.String(), nullable=False),
        sa.Column("property_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["privacy_experience_config_id"],
            ["privacyexperienceconfig.id"],
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["property_id"], ["plus_property.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("privacy_experience_config_id", "property_id"),
    )
    op.create_index(
        op.f("ix_plus_privacy_experience_config_property_privacy_experience_config_id"),
        "plus_privacy_experience_config_property",
        ["privacy_experience_config_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_plus_privacy_experience_config_property_property_id"),
        "plus_privacy_experience_config_property",
        ["property_id"],
        unique=False,
    )
    op.drop_index("ix_plus_property_id", table_name="plus_property")


def downgrade():
    op.create_index("ix_plus_property_id", "plus_property", ["id"], unique=False)
    op.drop_index(
        op.f("ix_plus_privacy_experience_config_property_property_id"),
        table_name="plus_privacy_experience_config_property",
    )
    op.drop_index(
        op.f("ix_plus_privacy_experience_config_property_privacy_experience_config_id"),
        table_name="plus_privacy_experience_config_property",
    )
    op.drop_table("plus_privacy_experience_config_property")



# pylint: skip-file
from json import dumps, loads
from typing import TYPE_CHECKING, Any, Dict, Type, TypeVar

from fides.api.util.custom_json_encoder import CustomJSONEncoder, _custom_decoder

if TYPE_CHECKING:
    # Override the Enum constructor not to accept a list of strings and confuse mypy:
    # https://github.com/dropbox/sqlalchemy-stubs/issues/114#issuecomment-541770319
    from sqlalchemy.sql.type_api import TypeEngine

    T = TypeVar("T")

    class EnumColumn(TypeEngine[T]):
        def __init__(self, enum: Type[T], **kwargs: Any) -> None: ...

else:
    from sqlalchemy import Enum as EnumColumn


def custom_json_deserializer(string: str) -> Dict:
    """
    JSON deserializer function to be used in SQLAlchemy engine creation.
    Leverages our custom JSON decoder function, which deserializes JSON
    data that's been serialized with the CustomJSONEncoder,
    to handle data types not supported by json.dumps() and json.loads()
    """
    return loads(string, object_hook=_custom_decoder)


def custom_json_serializer(obj: Any) -> str:
    """
    JSON serializer function to be used in SQLAlchemy engine creation.
    Leverages our CustomJSONEncoder, which can serialize data types that are
    not supported by json.dumps()
    """
    return dumps(obj, cls=CustomJSONEncoder)



from enum import Enum

from loguru import logger
from sqlalchemy.orm import Query, Session

from fides.api.common_exceptions import MessageDispatchException
from fides.api.models.policy import Policy, Rule
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.policy import ActionType, CurrentStep
from fides.api.schemas.privacy_request import PrivacyRequestStatus
from fides.api.service.connectors import get_connector
from fides.api.service.privacy_request.request_runner_service import (
    get_consent_email_connection_configs,
    get_erasure_email_connection_configs,
)
from fides.api.tasks import DatabaseTask, celery_app
from fides.api.tasks.scheduled.scheduler import scheduler
from fides.config import get_config
from fides.service.privacy_request.privacy_request_service import queue_privacy_request

CONFIG = get_config()
BATCH_EMAIL_SEND = "batch_email_send"


class EmailExitState(Enum):
    """A schema to describe where the email send process exited.
    For logging and testing"""

    no_applicable_privacy_requests = "no_applicable_privacy_requests"
    no_applicable_connectors = "no_applicable_connectors"
    missing_required_data = "missing_required_data"
    email_send_failed = "email_send_failed"
    complete = "complete"


@celery_app.task(base=DatabaseTask, bind=True)
def send_email_batch(self: DatabaseTask) -> EmailExitState:
    """Sends emails for each relevant connector with applicable user details batched together."""

    logger.info("Starting batched email send...")
    with self.get_new_session() as session:
        privacy_requests: Query = (
            session.query(PrivacyRequest)
            .filter(PrivacyRequest.status == PrivacyRequestStatus.awaiting_email_send)
            .filter(PrivacyRequest.deleted_at.is_(None))
            .order_by(PrivacyRequest.created_at.asc())  # oldest first
        )
        if not privacy_requests.first():
            logger.info(
                "Skipping batch email send with status: {}",
                EmailExitState.no_applicable_privacy_requests.value,
            )
            return EmailExitState.no_applicable_privacy_requests

        consent_configs: Query = get_consent_email_connection_configs(session)
        erasure_configs: Query = get_erasure_email_connection_configs(session)
        combined_configs = consent_configs.union_all(erasure_configs)
        if not combined_configs.first():
            requeue_privacy_requests_after_email_send(privacy_requests, session)
            logger.info(
                "Skipping batch email send with status: {}",
                EmailExitState.no_applicable_connectors.value,
            )
            return EmailExitState.no_applicable_connectors

        try:
            # erasure
            for connection_config in erasure_configs:
                get_connector(connection_config).batch_email_send(  # type: ignore
                    filter_privacy_requests_by_action_type(
                        privacy_requests, ActionType.erasure
                    )
                )

            # consent
            for connection_config in consent_configs:
                get_connector(connection_config).batch_email_send(  # type: ignore
                    filter_privacy_requests_by_action_type(
                        privacy_requests, ActionType.consent
                    )
                )
        except MessageDispatchException as exc:
            logger.error(
                "Batch email send for connector failed with exception: '{}'",
                exc,
            )
            return EmailExitState.email_send_failed

    requeue_privacy_requests_after_email_send(privacy_requests, session)
    return EmailExitState.complete


def filter_privacy_requests_by_action_type(
    privacy_requests: Query, action_type: ActionType
) -> Query:
    """Applies a filter for the specific Rule.action_type to the passed in privacy_requests."""
    return (
        privacy_requests.join(Policy, PrivacyRequest.policy_id == Policy.id)
        .join(Rule, Policy.id == Rule.policy_id)
        .filter(Rule.action_type == action_type)
    )


def requeue_privacy_requests_after_email_send(
    privacy_requests: Query, db: Session
) -> None:
    """After batch consent email send, requeue privacy requests from the post webhooks step
    to wrap up processing and transition to a "complete" state.

    Also cache on the privacy request itself that it is paused at the post-webhooks state,
    in case something happens in re-queueing.
    """
    logger.info("Batched email send complete.")
    logger.info("Queuing privacy requests from 'post_webhooks' step.")
    for privacy_request in privacy_requests:
        privacy_request.cache_paused_collection_details(
            step=CurrentStep.post_webhooks,
            collection=None,
            action_needed=None,
        )
        privacy_request.status = PrivacyRequestStatus.paused
        privacy_request.save(db=db)

        queue_privacy_request(
            privacy_request_id=privacy_request.id,
            from_step=CurrentStep.post_webhooks.value,
        )


def initiate_scheduled_batch_email_send() -> None:
    """Initiates scheduler to add weekly batch email send"""

    if CONFIG.test_mode:
        return

    assert scheduler.running, "Scheduler is not running! Cannot add Batch Email job."

    logger.info("Initiating scheduler for batch email send")
    scheduler.add_job(
        func=send_email_batch,
        kwargs={},
        id=BATCH_EMAIL_SEND,
        coalesce=False,
        replace_existing=True,
        trigger="cron",
        minute="0",
        hour="12",
        day_of_week="mon",
        timezone="US/Eastern",
    )



from typing import Dict

from loguru import logger

from fides.api.models.privacy_request import ProvidedIdentityType
from fides.api.schemas.drp_privacy_request import DrpIdentity
from fides.api.schemas.privacy_request import (
    PrivacyRequestDRPStatus,
    PrivacyRequestStatus,
)
from fides.api.schemas.redis_cache import Identity


class DrpFidesopsMapper:
    """
    Map DRP objects/enums to Fidesops
    """

    @staticmethod
    def map_identity(drp_identity: DrpIdentity) -> Identity:
        """
        Currently, both email and phone_number identity props map 1:1 to the corresponding
        Fidesops identity props in Identity. This may not always be the case.
        This class also allows us to implement custom logic to handle "verified" id props.
        """
        fidesops_identity_kwargs: Dict[str, str] = {}
        DRP_TO_FIDESOPS_SUPPORTED_IDENTITY_PROPS_MAP: Dict[
            str, ProvidedIdentityType
        ] = {
            "email": ProvidedIdentityType.email,
            "phone_number": ProvidedIdentityType.phone_number,
        }
        for attr, val in drp_identity.__dict__.items():
            if attr not in DRP_TO_FIDESOPS_SUPPORTED_IDENTITY_PROPS_MAP:
                logger.warning(
                    "Identity attribute of {} is not supported by Fidesops at this time. Continuing to use other identity props, if provided.",
                    attr,
                )
            else:
                fidesops_prop: str = DRP_TO_FIDESOPS_SUPPORTED_IDENTITY_PROPS_MAP[
                    attr
                ].value
                fidesops_identity_kwargs[fidesops_prop] = val
        return Identity(**fidesops_identity_kwargs)

    @staticmethod
    def map_status(
        status: PrivacyRequestStatus,
    ) -> PrivacyRequestDRPStatus:
        PRIVACY_REQUEST_STATUS_TO_DRP_MAPPING: Dict[
            PrivacyRequestStatus, PrivacyRequestDRPStatus
        ] = {
            PrivacyRequestStatus.pending: PrivacyRequestDRPStatus.open,
            PrivacyRequestStatus.approved: PrivacyRequestDRPStatus.in_progress,
            PrivacyRequestStatus.denied: PrivacyRequestDRPStatus.denied,
            PrivacyRequestStatus.in_processing: PrivacyRequestDRPStatus.in_progress,
            PrivacyRequestStatus.complete: PrivacyRequestDRPStatus.fulfilled,
            PrivacyRequestStatus.paused: PrivacyRequestDRPStatus.in_progress,
            PrivacyRequestStatus.error: PrivacyRequestDRPStatus.expired,
            PrivacyRequestStatus.canceled: PrivacyRequestDRPStatus.revoked,
        }
        try:
            return PRIVACY_REQUEST_STATUS_TO_DRP_MAPPING[status]
        except KeyError:
            raise ValueError(f"Request has invalid DRP request status: {status.value}")



import hashlib
import hmac
from typing import Callable

from fides.api.schemas.masking.masking_configuration import HmacMaskingConfiguration
from fides.config import CONFIG


def hmac_encrypt_return_bytes(
    value: str,
    hmac_key: str,
    salt: str,
    hashing_algorithm: HmacMaskingConfiguration.Algorithm,
) -> bytes:
    return _hmac_encrypt(value, hmac_key, salt, hashing_algorithm).digest()


def hmac_encrypt_return_str(
    value: str,
    hmac_key: str,
    salt: str,
    hashing_algorithm: HmacMaskingConfiguration.Algorithm,
) -> str:
    return _hmac_encrypt(value, hmac_key, salt, hashing_algorithm).hexdigest()


def _hmac_encrypt(
    value: str,
    hmac_key: str,
    salt: str,
    hashing_algorithm: HmacMaskingConfiguration.Algorithm,
) -> hmac.HMAC:
    """Generic HMAC algorithm"""

    algorithm_function_mapping = {
        HmacMaskingConfiguration.Algorithm.sha_256: _hmac_sha256,
        HmacMaskingConfiguration.Algorithm.sha_512: _hmac_sha512,
    }

    if hashing_algorithm not in algorithm_function_mapping:
        raise ValueError(f"{hashing_algorithm} is an unsupported hashing_algorithm")

    algorithm_function = algorithm_function_mapping[hashing_algorithm]
    return algorithm_function(value, hmac_key, salt)


def _hmac_sha256(value: str, hmac_key: str, salt: str) -> hmac.HMAC:
    """Creates a new hmac object using the sh256 hash algorithm and the hmac_key and then returns the hexdigest."""
    return _hmac(value=value, hmac_key=hmac_key, salt=salt, hashing_alg=hashlib.sha256)


def _hmac_sha512(value: str, hmac_key: str, salt: str) -> hmac.HMAC:
    """Creates a new hmac object using the sha512 hash algorithm and the hmac_key and then returns the hexdigest."""
    return _hmac(value=value, hmac_key=hmac_key, salt=salt, hashing_alg=hashlib.sha512)


def _hmac(value: str, hmac_key: str, salt: str, hashing_alg: Callable) -> hmac.HMAC:
    return hmac.new(
        key=hmac_key.encode(CONFIG.security.encoding),
        msg=(value + salt).encode(CONFIG.security.encoding),
        digestmod=hashing_alg,
    )



from __future__ import annotations

"""Update fideslang data categories

Revision ID: 7abe778b7082
Revises: bab75915670a
Create Date: 2022-07-29 17:54:53.719453

"""
import logging

from sqlalchemy.exc import ProgrammingError

from fides.api.db.base import DatasetConfig
from fides.api.db.session import get_db_session
from fides.config import CONFIG

logger = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = "7abe778b7082"
down_revision = "bab75915670a"
branch_labels = None
depends_on = None


DATA_MAP_UPGRADE = {
    "account.payment": "user.financial.payment_information",
    "user.derived.identifiable.biometric_health": "user.biometric_health",
    "user.derived.identifiable.browsing_history": "user.browsing_history",
    "user.derived.identifiable.demographic": "user.demographic",
    "user.derived.identifiable.device": "user.device",
    "user.derived.identifiable.device.cookie_id": "user.device.cookie_id",
    "user.derived.identifiable.device.device_id": "user.device.device_id",
    "user.derived.identifiable.device.ip_address": "user.device.ip_address",
    "user.derived.identifiable.gender": "user.gender",
    "user.derived.identifiable.location": "user.location",
    "user.derived.identifiable.media_consumption": "user.media_consumption",
    "user.derived.identifiable.non_specific_age": "user.non_specific_age",
    "user.derived.identifiable.observed": "user.observed",
    "user.derived.identifiable.profiling": "user.profiling",
    "user.derived.identifiable.race": "user.race",
    "user.derived.identifiable.religious_belief": "user.religious_belief",
    "user.derived.identifiable.search_history": "user.search_history",
    "user.derived.identifiable.sexual_orientation": "user.sexual_orientation",
    "user.derived.identifiable.social": "user.social",
    "user.derived.identifiable.telemetry": "user.telemetry",
    "user.derived.identifiable.unique_id": "user.unique_id",
    "user.derived.identifiable.user_sensor": "user.sensor_data",
    "user.derived.identifiable.organization": "user.organization",
    "user.derived.identifiable.workplace": "user.workplace",
    "user.provided.identifiable.biometric": "user.biometric",
    "user.provided.identifiable.childrens": "user.childrens",
    "user.provided.identifiable.contact.city": "user.contact.address.city",
    "user.provided.identifiable.contact.country": "user.contact.address.country",
    "user.provided.identifiable.contact.postal_code": "user.contact.address.postal_code",
    "user.provided.identifiable.contact.state": "user.contact.address.state",
    "user.provided.identifiable.contact.street": "user.contact.address.street",
    "user.provided.identifiable.contact.email": "user.contact.email",
    "user.provided.identifiable.contact.phone_number": "user.contact.phone_number",
    "user.provided.identifiable.credentials": "user.credentials",
    "user.provided.identifiable.credentials.biometric_credentials": "user.credentials.biometric_credentials",
    "user.provided.identifiable.credentials.password": "user.credentials.password",
    "user.provided.identifiable.date_of_birth": "user.date_of_birth",
    "user.provided.identifiable.financial": "user.financial",
    "user.provided.identifiable.financial.account_number": "user.financial.account_number",
    "user.provided.identifiable.gender": "user.gender",
    "user.provided.identifiable.genetic": "user.genetic",
    "user.provided.identifiable.government_id": "user.government_id",
    "user.provided.identifiable.government_id.drivers_license_number": "user.government_id.drivers_license_number",
    "user.provided.identifiable.government_id.national_identification_number": "user.government_id.national_identification_number",
    "user.provided.identifiable.government_id.passport_number": "user.government_id.passport_number",
    "user.provided.identifiable.health_and_medical": "user.health_and_medical",
    "user.provided.identifiable.job_title": "user.job_title",
    "user.provided.identifiable.name": "user.name",
    "user.provided.identifiable.non_specific_age": "user.non_specific_age",
    "user.provided.identifiable.political_opinion": "user.political_opinion",
    "user.provided.identifiable.race": "user.race",
    "user.provided.identifiable.religious_belief": "user.religious_belief",
    "user.provided.identifiable.sexual_orientation": "user.sexual_orientation",
    "user.provided.identifiable.workplace": "user.provided.identifiable.workplace",
}

DATA_MAP_DOWNGRADE = {v: k for k, v in DATA_MAP_UPGRADE.items()}


def is_removed_downgrade(privacy_key: str) -> bool:
    removed = (
        "user.contact.address",
        "user.contact.address.number",
        "user.financial.payment_information",
        "user.name.first",
        "user.name.last",
    )

    return privacy_key in removed


def is_removed_upgrade(privacy_key: str) -> bool:
    removed = (
        "account",
        "account.contact",
        "account.contact.city",
        "account.contact.country",
        "account.contact.email",
        "account.contact.phone_number",
        "account.contact.postal_code",
        "account.contact.state",
        "account.contact.street",
        "account.payment.financial_account_number",
        "user.derived",
        "user.derived.identifiable",
        "user.derived.nonidentifiable",
        "user.derived.nonidentifiable.sensor",
        "user.provided",
        "user.provided.identifiable",
        "user.provided.nonidentifiable",
    )

    return privacy_key in removed


def update_field(field: dict, migration_direction: str) -> dict:
    data_categories = field.get("data_categories", [])
    updated = []
    if migration_direction == "up":
        for data_category in data_categories:
            if is_removed_upgrade(data_category):
                logger.info(
                    "Removing %s, this is no longer a valid category", data_category
                )
            if data_category:
                if not is_removed_upgrade(data_category):
                    new = DATA_MAP_UPGRADE.get(data_category)
                    if new:
                        updated.append(new)
                    else:
                        updated.append(data_category)
                else:
                    updated.append(data_category)
        field["data_categories"] = updated
    else:
        for data_category in data_categories:
            if is_removed_downgrade(data_category):
                logger.info(
                    "Removing %s, this is no longer a valid category", data_category
                )
            if data_category:
                if not is_removed_downgrade(data_category):
                    new = DATA_MAP_DOWNGRADE.get(data_category)
                    if new:
                        updated.append(new)
                    else:
                        updated.append(data_category)
                else:
                    updated.append(data_category)
        field["data_categories"] = updated

    # Document databases like mongo can have nested fields so this is the reason
    # for the recursion here.
    if field.get("fields"):
        for field in field["fields"]:
            update_field(field, migration_direction)

    return field


def run_migration(migration_direction: str) -> None:
    sessionlocal = get_db_session(CONFIG)
    with sessionlocal() as session:
        try:
            datasets = DatasetConfig.all(db=session)

            logger.info("Upgrading fideslang categories")
            counter = 0
            for dataset in datasets:
                counter += 1
                if counter % 10 == 0:
                    logger.info("Upgrading fideslang categories")

                dataset_json = dataset.dataset
                collections = dataset_json.get("collections")

                if collections:
                    for collection in collections:
                        fields = collection.get("fields")
                        if fields:
                            for field in fields:
                                field = update_field(field, migration_direction)

        # If the datasetconfig table doesn't exist a ProgrammingError will be thrown.
        # If that happens we know there are no records to update so we can ignore the
        # error.
        except ProgrammingError:
            pass


def upgrade() -> None:
    run_migration("up")


def downgrade() -> None:
    run_migration("down")



"""add role based permissions

Revision ID: eb1e6ec39b83
Revises: d65bbc647083
Create Date: 2023-02-24 19:27:17.844231

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
from sqlalchemy.dialects import postgresql

revision = "eb1e6ec39b83"
down_revision = "d65bbc647083"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "client",
        sa.Column("roles", sa.ARRAY(sa.String()), server_default="{}", nullable=False),
    )
    op.add_column(
        "fidesuserpermissions",
        sa.Column("roles", sa.ARRAY(sa.String()), server_default="{}", nullable=False),
    )
    op.alter_column(
        "fidesuserpermissions",
        "scopes",
        existing_type=postgresql.ARRAY(sa.VARCHAR()),
        nullable=False,
        server_default="{}",
    )
    op.alter_column(
        "client",
        "scopes",
        existing_type=postgresql.ARRAY(sa.VARCHAR()),
        nullable=False,
        server_default="{}",
    )


def downgrade():
    op.drop_column("fidesuserpermissions", "roles")
    op.drop_column("client", "roles")

    op.alter_column(
        "fidesuserpermissions",
        "scopes",
        existing_type=postgresql.ARRAY(sa.VARCHAR()),
        nullable=False,
        server_default=None,
    )

    op.alter_column(
        "client",
        "scopes",
        existing_type=postgresql.ARRAY(sa.VARCHAR()),
        nullable=False,
        server_default=None,
    )



"""Change Attachment.storage_key to foreign key

Revision ID: 6ea2171c544f
Revises: 1152c1717849
Create Date: 2025-03-10 17:36:31.504831

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "6ea2171c544f"
down_revision = "1152c1717849"
branch_labels = None
depends_on = None


def upgrade():

    # Alter the column type and add the new foreign key constraint
    op.alter_column(
        "attachment",
        "storage_key",
        existing_type=sa.String(),
        type_=sa.String(),
        nullable=False,
    )
    op.create_foreign_key(
        "fk_attachment_storage_key",
        "attachment",
        "storageconfig",
        ["storage_key"],
        ["key"],
        ondelete="CASCADE",
    )

    # Add index on attachment_reference.reference_id
    op.create_index(
        "ix_attachment_reference_reference_id",
        "attachment_reference",
        ["reference_id"],
    )

    # Add index on attachment_reference.reference_type
    op.create_index(
        "ix_attachment_reference_reference_type",
        "attachment_reference",
        ["reference_type"],
    )
    # ### end Alembic commands ###


def downgrade():
    # Drop the index on attachment_reference.reference_id
    op.drop_index(
        "ix_attachment_reference_reference_id", table_name="attachment_reference"
    )

    # Drop the index on attachment_reference.reference_type
    op.drop_index(
        "ix_attachment_reference_reference_type", table_name="attachment_reference"
    )

    # Drop the new foreign key constraint
    op.drop_constraint("fk_attachment_storage_key", "attachment", type_="foreignkey")

    # Revert the column type change
    op.alter_column(
        "attachment",
        "storage_key",
        existing_type=sa.String(),
        type_=sa.String(),
        nullable=True,
    )
    # ### end Alembic commands ###



import inspect
from abc import abstractmethod
from enum import Enum
from functools import wraps
from typing import Any, Callable, Dict, Optional

from loguru import logger
from requests import PreparedRequest, Response
from requests.exceptions import (  # pylint: disable=redefined-builtin
    ConnectionError,
    ConnectTimeout,
    ReadTimeout,
    SSLError,
    TooManyRedirects,
)

from fides.config import CONFIG


class LoggerContextKeys(str, Enum):
    action_type = "action_type"
    status_code = "status_code"
    body = "body"
    collection = "collection"
    connection_key = "connection_key"
    error_details = "error_details"
    error_group = "error_group"
    method = "method"
    privacy_request_id = "privacy_request_id"
    response = "response"
    system_key = "system_key"
    url = "url"
    task_id = "task_id"
    privacy_request_source = "privacy_request_source"


class ErrorGroup(str, Enum):
    """A collection of user-friendly error labels to be used in contextualized logs."""

    network_error = "NetworkError"
    authentication_error = "AuthenticationError"
    client_error = "ClientSideError"
    server_error = "ServerSideError"


class Contextualizable:
    """
    An abstract base class that defines a contract for classes which can provide
    contextual information for logging purposes.

    Subclasses of Contextualizable must implement the get_log_context method,
    which should return a dictionary of context information relevant to the object.
    This context will be used by the log_context decorator to add additional
    information to log messages.
    """

    @abstractmethod
    def get_log_context(self) -> Dict[LoggerContextKeys, Any]:
        pass


def log_context(
    _func: Optional[Callable] = None,
    capture_args: Optional[dict[str, LoggerContextKeys]] = None,
    **additional_context: Any,
) -> Callable:
    """
    A decorator that adds context information to log messages. It extracts:
    1. Values from function arguments, mapping them to standardized context names via capture_args dict
    2. Context from any Contextualizable arguments
    3. Additional context provided as decorator kwargs

    Example:
       @log_context(capture_args={"user_id": "standard_user_id"}, tenant="example")
       def process_user(user_id: str) -> None:

    Logs will include standard_user_id=<user_id> and tenant="example"
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            context = dict(additional_context)

            # extract specified param values from kwargs and args
            if capture_args:
                # First, process kwargs as they're explicitly named
                for arg_name, context_name in capture_args.items():
                    if arg_name in kwargs:
                        context[context_name.value] = kwargs[arg_name]

                # Process args using signature binding for more robust parameter mapping
                if args:
                    try:
                        # Get the signature and bind the arguments
                        sig = inspect.signature(func)
                        # This will map positional args to their parameter names correctly
                        bound_args = sig.bind_partial(*args, **kwargs)

                        # Now we can iterate through the bound arguments
                        for param_name, arg_value in bound_args.arguments.items():
                            # Only process if this parameter is in capture_args and wasn't already found in kwargs
                            if param_name in capture_args and param_name not in kwargs:
                                context_name = capture_args[param_name]
                                context[context_name.value] = arg_value
                    except TypeError:
                        # Handle the case where the arguments don't match the signature
                        pass

                # Handle default parameters that weren't provided in args or kwargs
                if capture_args:
                    sig = inspect.signature(func)
                    for param_name, param in sig.parameters.items():
                        # Check if parameter has a default value and is in capture_args
                        # and hasn't been processed yet (not in context)
                        if (
                            param.default is not param.empty
                            and param_name in capture_args
                            and capture_args[param_name].value not in context
                        ):
                            context_name = capture_args[param_name]
                            context[context_name.value] = param.default

            # Process Contextualizable args
            for arg in args:
                if isinstance(arg, Contextualizable):
                    arg_context = arg.get_log_context()
                    # get values from enums
                    context.update(
                        {
                            key.value if isinstance(key, Enum) else key: value
                            for key, value in arg_context.items()
                        }
                    )
            with logger.contextualize(**context):
                return func(*args, **kwargs)

        return wrapper

    if _func is None:
        return decorator
    return decorator(_func)


def request_details(
    prepared_request: PreparedRequest,
    response: Optional[Response] = None,
    ignore_error: bool = False,
) -> Dict[str, Any]:
    """
    Maps the request details and includes response details when "dev mode" is enabled.
    """

    details: Dict[str, Any] = {
        LoggerContextKeys.method.value: prepared_request.method,
        LoggerContextKeys.url.value: prepared_request.url,
    }
    if CONFIG.dev_mode and prepared_request.body is not None:
        if isinstance(prepared_request.body, bytes):
            details[LoggerContextKeys.body.value] = prepared_request.body.decode(
                "utf-8"
            )
        elif isinstance(prepared_request.body, str):
            details[LoggerContextKeys.body.value] = prepared_request.body

    if response is not None:
        if CONFIG.dev_mode and response.content:
            details[LoggerContextKeys.response.value] = response.content.decode("utf-8")

        details[LoggerContextKeys.status_code.value] = response.status_code

        # assign error group only if error should not be ignored
        if not ignore_error:
            if response.status_code in [401, 403]:
                details[LoggerContextKeys.error_group.value] = (
                    ErrorGroup.authentication_error.value
                )
            elif 400 <= response.status_code < 500:
                details[LoggerContextKeys.error_group.value] = (
                    ErrorGroup.client_error.value
                )
            elif 500 <= response.status_code:
                details[LoggerContextKeys.error_group.value] = (
                    ErrorGroup.server_error.value
                )
    return details


def connection_exception_details(exception: Exception, url: str) -> Dict[str, Any]:
    """Maps select connection exceptions to user-friendly error details."""

    details = {
        LoggerContextKeys.error_group.value: ErrorGroup.network_error.value,
        LoggerContextKeys.error_details.value: f"Unknown exception connecting to {url}.",
        LoggerContextKeys.status_code.value: None,
    }
    if isinstance(exception, ConnectTimeout):
        details[LoggerContextKeys.error_details.value] = (
            f"Timeout occurred connecting to {url}."
        )
    elif isinstance(exception, ReadTimeout):
        details[LoggerContextKeys.error_details.value] = (
            f"Timeout occurred waiting for a response from {url}."
        )
    elif isinstance(exception, SSLError):
        details[LoggerContextKeys.error_details.value] = (
            f"SSL exception occurred connecting to {url}."
        )
    elif isinstance(exception, TooManyRedirects):
        details[LoggerContextKeys.error_details.value] = (
            f"Too many redirects occurred connecting to {url}."
        )
    elif isinstance(exception, ConnectionError):
        details[LoggerContextKeys.error_details.value] = f"Unable to connect to {url}."
    return details



"""Add plus_system_history table

Revision ID: 093bb28a8270
Revises: 3038667ba898
Create Date: 2023-08-18 23:48:22.934916

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "093bb28a8270"
down_revision = "3038667ba898"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "plus_system_history",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("user_id", sa.String(), nullable=True),
        sa.Column("system_id", sa.String(), nullable=False),
        sa.Column("before", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column("after", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.ForeignKeyConstraint(["system_id"], ["ctl_systems.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_plus_system_history_id"), "plus_system_history", ["id"], unique=False
    )
    op.create_index(
        "idx_plus_system_history_created_at_system_id",
        "plus_system_history",
        ["created_at", "system_id"],
    )
    op.add_column(
        "ctl_systems",
        sa.Column("user_id", sa.String, nullable=True),
    )


def downgrade():
    op.drop_column("ctl_systems", "user_id")
    op.drop_index(
        op.f("idx_plus_system_history_created_at_system_id"),
        table_name="plus_system_history",
    )
    op.drop_index(op.f("ix_plus_system_history_id"), table_name="plus_system_history")
    op.drop_table("plus_system_history")



"""
Defines the logging format and other helper functions to be used throughout the API server code.
"""

# pylint: disable=eval-used,no-member

from __future__ import annotations

import logging
import sys
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional

from loguru import logger
from loguru._handler import Message

from fides.api.schemas.privacy_request import LogEntry, PrivacyRequestSource
from fides.config import CONFIG, FidesConfig

if TYPE_CHECKING:
    from fides.api.util.cache import FidesopsRedis

MASKED = "MASKED"


class RedisSink:
    """A sink that writes log messages to Redis."""

    def __init__(self) -> None:
        self.cache: Optional["FidesopsRedis"] = None

    def _ensure_cache(self) -> None:
        """Lazily initialize Redis connection when needed."""
        if self.cache is None:
            from fides.api.util.cache import get_cache

            self.cache = get_cache()

    def __call__(self, message: Message) -> None:
        """Write log message to Redis if conditions are met."""

        record: Dict[str, Any] = message.record  # type: ignore[attr-defined]

        # Extract privacy request context
        extras = record["extra"]
        privacy_request_id = extras.get("privacy_request_id")
        privacy_request_source = extras.get("privacy_request_source")

        # Only process logs with privacy request ID and source is dataset_test
        if (
            not privacy_request_id
            or privacy_request_source != PrivacyRequestSource.dataset_test.value
        ):
            return

        # Ensure we have a Redis connection
        self._ensure_cache()
        assert self.cache  # for mypy

        # Create Redis key using privacy request ID
        key = f"log_{privacy_request_id}"

        # Format log message
        module_info = f"{record['name']}:{record['function']}:{record['line']}"

        log_entry = LogEntry(
            timestamp=record["time"].isoformat(),
            level=record["level"].name,
            module_info=module_info,
            message=record["message"],
        )

        # Encode and append log entry to the list in Redis
        self.cache.push_encoded_object(key, log_entry, expire_time=43200)  # 12 hours


class Pii(str):
    """Mask pii data"""

    def __format__(self, __format_spec: str) -> str:
        if CONFIG.logging.log_pii:
            return super().__format__(__format_spec)
        return MASKED


def _log_exception(exc: BaseException, dev_mode: bool = False) -> None:
    """If dev mode, log the entire traceback"""
    if dev_mode:
        logger.opt(exception=True).error(exc)
    else:
        logger.error(exc)


def _log_warning(exc: BaseException, dev_mode: bool = False) -> None:
    """If dev mode, log the entire traceback"""
    if dev_mode:
        logger.opt(exception=True).warning(exc)
    else:
        logger.error(exc)


def create_handler_dicts(
    level: str, sink: Any, serialize: bool, colorize: bool, include_called_from: bool
) -> List[Dict]:
    """Creates dictionaries used for configuring loguru handlers."""
    time_format = "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green>"
    level_format = "<level>{level: <8}</level>"
    called_from_format = (
        ("<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan>")
        if include_called_from
        else ""
    )
    message_format = "<level>{message}</level>"

    log_format = (
        time_format
        + " | "
        + level_format
        + " | "
        + called_from_format
        + " - "
        + message_format
    )

    # Create base handler config
    base_config = {
        "format": log_format,
        "level": level,
        "serialize": serialize,
        "colorize": colorize,
        "diagnose": False,
        "backtrace": True,
        "catch": True,
    }

    # Configure handler
    standard_dict = {
        **base_config,
        "sink": sink,
        "filter": lambda logRecord: not bool(logRecord["extra"]),
    }

    # Create extra dict with additional formatting for logs with extra context
    extra_dict = {
        **standard_dict,
        "format": log_format + " | <dim>{extra}</dim>",
        "filter": lambda logRecord: bool(logRecord["extra"]),
    }

    return [standard_dict, extra_dict]


def setup(config: FidesConfig) -> None:
    """
    Configures logging with the appropriate sink based on configuration.
    Removes all handlers from all loggers, and sets those loggers to propagate
    log entries to the root logger. Then, configures Loguru to use the desired handlers.
    """
    for name in logging.root.manager.loggerDict.keys():
        logging.getLogger(name).handlers = []
        logging.getLogger(name).propagate = True

    handlers = []

    # Configure main sink from config
    destination = config.logging.destination
    main_sink = sys.stdout if not destination else destination
    handlers.extend(
        create_handler_dicts(
            level=config.logging.level,
            include_called_from=config.dev_mode,
            sink=main_sink,
            serialize=config.logging.serialization == "json",
            colorize=config.logging.colorize,
        )
    )

    # Add Redis sink if Redis is enabled
    if config.redis.enabled:
        redis_sink = RedisSink()
        handlers.extend(
            create_handler_dicts(
                level=config.logging.level,
                include_called_from=config.dev_mode,
                sink=redis_sink,
                serialize=True,  # Always serialize for Redis
                colorize=False,  # Redis doesn't need colorization
            )
        )

    logger.configure(handlers=handlers)


def obfuscate_message(message: str) -> str:
    """
    Obfuscate specific bits of information that might get logged.

    Currently being obfuscated:
        - Database username + password
    """
    strings_to_replace: List[str] = [
        f"{CONFIG.database.user}:{CONFIG.database.password}"
    ]

    for string in strings_to_replace:
        result = message.replace(string, "*****")

    return result


# Loguru doesn't export the Record type so this can't be properly typed
# Taken from the following issue:
# https://github.com/Delgan/loguru/issues/537#issuecomment-986259036
def format_and_obfuscate(record) -> str:  # type: ignore[no-untyped-def]
    record["extra"]["obfuscated_message"] = obfuscate_message(record["message"])
    return "[{level}] {extra[obfuscated_message]}\n{exception}"


# pylint: disable=protected-access
@contextmanager
def suppress_logging() -> Generator:
    """
    Temporarily suppress all logging by setting the log level to CRITICAL.
    """

    current_level = logger._core.min_level  # type: ignore[attr-defined]
    logger._core.min_level = logger.level("CRITICAL").no  # type: ignore[attr-defined]

    try:
        yield
    finally:
        # Restore original configuration
        logger._core.min_level = current_level  # type: ignore[attr-defined]



from __future__ import annotations

from typing import Any, Dict

from loguru import logger
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.engine.url import URL
from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy.pool import NullPool

from fides.api.common_exceptions import MissingConfig
from fides.api.db.util import custom_json_deserializer, custom_json_serializer
from fides.config import FidesConfig


def get_db_engine(
    *,
    config: FidesConfig | None = None,
    database_uri: str | URL | None = None,
    pool_size: int = 50,
    max_overflow: int = 50,
    keepalives_idle: int | None = None,
    keepalives_interval: int | None = None,
    keepalives_count: int | None = None,
    disable_pooling: bool = False,
) -> Engine:
    """Return a database engine.

    If the TESTING environment var is set the database engine returned will be
    connected to the test DB.
    """
    if not config and not database_uri:
        raise ValueError("Either a config or database_uri is required")

    if not database_uri and config:
        # Don't override any database_uri explicitly passed in
        if config.test_mode:
            database_uri = config.database.sqlalchemy_test_database_uri
        else:
            database_uri = config.database.sqlalchemy_database_uri

    engine_args: Dict[str, Any] = {
        "json_serializer": custom_json_serializer,
        "json_deserializer": custom_json_deserializer,
    }

    # keepalives settings
    connect_args = {}
    if keepalives_idle:
        connect_args["keepalives_idle"] = keepalives_idle
    if keepalives_interval:
        connect_args["keepalives_interval"] = keepalives_interval
    if keepalives_count:
        connect_args["keepalives_count"] = keepalives_count

    if connect_args:
        connect_args["keepalives"] = 1
        engine_args["connect_args"] = connect_args

    if disable_pooling:
        engine_args["poolclass"] = NullPool
    else:
        engine_args["pool_pre_ping"] = True
        engine_args["pool_size"] = pool_size
        engine_args["max_overflow"] = max_overflow

    return create_engine(database_uri, **engine_args)


def get_db_session(
    config: FidesConfig,
    autocommit: bool = False,
    autoflush: bool = False,
    engine: Engine | None = None,
) -> sessionmaker:
    """Return a database SessionLocal."""
    if not config.database.sqlalchemy_database_uri:
        raise MissingConfig("No database uri available in the config")

    return sessionmaker(
        autocommit=autocommit,
        autoflush=autoflush,
        bind=engine or get_db_engine(config=config),
        class_=ExtendedSession,
    )


class ExtendedSession(Session):
    """This class wraps the SQLAlchemy Session to provide some error handling on
    commits."""

    def commit(self) -> None:
        """Provide the option to automatically rollback failed transactions."""
        try:
            return super().commit()
        except Exception as exc:
            logger.error("Exception: {}", exc)
            # Rollback the current transaction after each failed commit
            self.rollback()
            raise



"""record consent preferences

Revision ID: 39a23bc016ea
Revises: a0f219697fa0
Create Date: 2023-03-28 02:42:13.114938

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "39a23bc016ea"
down_revision = "a0f219697fa0"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "consentrequest",
        sa.Column(
            "preferences", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
    )
    op.add_column(
        "consentrequest",
        sa.Column("identity_verified_at", sa.DateTime(timezone=True), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("consentrequest", "identity_verified_at")
    op.drop_column("consentrequest", "preferences")
    # ### end Alembic commands ###



from __future__ import annotations

from typing import Annotated, Any, Dict, List, Optional

from fastapi import Depends
from fastapi.params import Query, Security
from fastapi_pagination import Page, Params
from fastapi_pagination.bases import AbstractPage
from fastapi_pagination.ext.sqlalchemy import paginate
from fideslang.validation import FidesKey
from loguru import logger
from pydantic import Field
from sqlalchemy import null, or_
from sqlalchemy.orm import Session
from sqlalchemy_utils import escape_like
from starlette.status import HTTP_200_OK, HTTP_204_NO_CONTENT

from fides.api.api import deps
from fides.api.models.connectionconfig import ConnectionConfig, ConnectionType
from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas.connection_configuration import connection_secrets_schemas
from fides.api.schemas.connection_configuration.connection_config import (
    BulkPutConnectionConfiguration,
    ConnectionConfigurationResponse,
    CreateConnectionConfigurationWithSecrets,
)
from fides.api.schemas.connection_configuration.connection_secrets import (
    TestStatusMessage,
)
from fides.api.schemas.connection_configuration.enums.system_type import SystemType
from fides.api.schemas.connection_configuration.enums.test_status import TestStatus
from fides.api.util.api_router import APIRouter
from fides.api.util.connection_util import (
    connection_status,
    delete_connection_config,
    get_connection_config_or_error,
    patch_connection_configs,
    validate_secrets,
)
from fides.common.api.scope_registry import (
    CONNECTION_CREATE_OR_UPDATE,
    CONNECTION_DELETE,
    CONNECTION_READ,
)
from fides.common.api.v1.urn_registry import (
    CONNECTION_BY_KEY,
    CONNECTION_SECRETS,
    CONNECTION_TEST,
    CONNECTIONS,
    V1_URL_PREFIX,
)

router = APIRouter(tags=["Connections"], prefix=V1_URL_PREFIX)


@router.get(
    CONNECTIONS,
    dependencies=[Security(verify_oauth_client, scopes=[CONNECTION_READ])],
    response_model=Page[ConnectionConfigurationResponse],
)
def get_connections(
    *,
    db: Session = Depends(deps.get_db),
    params: Params = Depends(),
    search: Optional[str] = None,
    disabled: Optional[bool] = None,
    test_status: Optional[TestStatus] = None,
    system_type: Optional[SystemType] = None,
    orphaned_from_system: Optional[bool] = None,
    connection_type: Optional[List[str]] = Query(default=None),  # type: ignore
) -> AbstractPage[ConnectionConfig]:
    # pylint: disable=too-many-branches
    """Returns all connection configurations in the database.
    Optionally filter the key, name, and description with a search query param.

    Can also filter on disabled, connection_type, test_status, and system_type.

    Connection_type supports "or" filtering:
    ?connection_type=postgres&connection_type=mongo will be translated
    into an "or" query. This parameter can also be used to filter by specific
    SaaS connector types.
    """
    logger.info(
        "Finding connection configurations with pagination params {} and search query: '{}'.",
        params,
        search if search else "",
    )
    query = ConnectionConfig.query(db)

    if search:
        query = query.filter(
            or_(
                ConnectionConfig.key.ilike(f"%{escape_like(search)}%"),
                ConnectionConfig.name.ilike(f"%{escape_like(search)}%"),
                ConnectionConfig.description.ilike(f"%{escape_like(search)}%"),
            )
        )

    if connection_type:
        connection_types = []
        saas_connection_types = []
        for ct in connection_type:
            ct = ct.lower()
            try:
                conn_type = ConnectionType(ct)
                connection_types.append(conn_type)
            except ValueError:
                # if not a ConnectionType enum, assume it's
                # a SaaS type, since those are dynamic
                saas_connection_types.append(ct)
        query = query.filter(
            or_(
                ConnectionConfig.connection_type.in_(connection_types),
                ConnectionConfig.saas_config["type"].astext.in_(saas_connection_types),
            )
        )

    if disabled is not None:
        query = query.filter(ConnectionConfig.disabled == disabled)

    if test_status:
        query = query.filter(
            ConnectionConfig.last_test_succeeded.is_(test_status.str_to_bool())
        )

    if orphaned_from_system is not None:
        if orphaned_from_system:
            query = query.filter(ConnectionConfig.system_id.is_(null()))
        else:
            query = query.filter(ConnectionConfig.system_id.is_not(null()))  # type: ignore

    if system_type:
        if system_type == SystemType.saas:
            query = query.filter(
                ConnectionConfig.connection_type == ConnectionType.saas
            )
        elif system_type == SystemType.manual:
            query = query.filter(
                ConnectionConfig.connection_type == ConnectionType.manual
            )
        elif system_type == SystemType.database:
            query = query.filter(
                ConnectionConfig.connection_type.notin_(
                    [ConnectionType.saas, ConnectionType.manual]
                )
            )

    return paginate(
        query.order_by(ConnectionConfig.name.asc()),
        params=params,
    )


@router.get(
    CONNECTION_BY_KEY,
    dependencies=[Security(verify_oauth_client, scopes=[CONNECTION_READ])],
    response_model=ConnectionConfigurationResponse,
)
def get_connection_detail(
    connection_key: FidesKey, db: Session = Depends(deps.get_db)
) -> ConnectionConfig:
    """Returns connection configuration with matching key."""
    return get_connection_config_or_error(db, connection_key)


@router.patch(
    CONNECTIONS,
    dependencies=[Security(verify_oauth_client, scopes=[CONNECTION_CREATE_OR_UPDATE])],
    status_code=HTTP_200_OK,
    response_model=BulkPutConnectionConfiguration,
)
def patch_connections(
    *,
    db: Session = Depends(deps.get_db),
    configs: Annotated[List[CreateConnectionConfigurationWithSecrets], Field(max_length=50)],  # type: ignore
) -> BulkPutConnectionConfiguration:
    """
    Given a list of connection config data elements, optionally containing the secrets,
    create or update corresponding ConnectionConfig objects or report failure

    If the key in the payload exists, it will be used to update an existing ConnectionConfiguration.
    Otherwise, a new ConnectionConfiguration will be created for you.
    """
    return patch_connection_configs(db, configs)


@router.delete(
    CONNECTION_BY_KEY,
    dependencies=[Security(verify_oauth_client, scopes=[CONNECTION_DELETE])],
    status_code=HTTP_204_NO_CONTENT,
    response_model=None,
)
def delete_connection(
    connection_key: FidesKey, *, db: Session = Depends(deps.get_db)
) -> None:
    delete_connection_config(db, connection_key)


def validate_and_update_secrets(
    connection_key: FidesKey,
    connection_config: ConnectionConfig,
    db: Session,
    unvalidated_secrets: connection_secrets_schemas,
    verify: Optional[bool],
) -> TestStatusMessage:
    connection_config.secrets = validate_secrets(
        db, unvalidated_secrets, connection_config
    ).model_dump(mode="json")
    # Save validated secrets, regardless of whether they've been verified.
    logger.info("Updating connection config secrets for '{}'", connection_key)
    connection_config.save(db=db)

    msg = f"Secrets updated for ConnectionConfig with key: {connection_key}."

    if verify:
        return connection_status(connection_config, msg, db)

    return TestStatusMessage(msg=msg, test_status=None)


@router.put(
    CONNECTION_SECRETS,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[CONNECTION_CREATE_OR_UPDATE])],
    response_model=TestStatusMessage,
)
def put_connection_config_secrets(
    connection_key: FidesKey,
    *,
    db: Session = Depends(deps.get_db),
    unvalidated_secrets: connection_secrets_schemas,
    verify: Optional[bool] = True,
) -> TestStatusMessage:
    """
    Update secrets that will be used to connect to a specified connection_type.

    The specific secrets will be connection-dependent. For example, the components needed to connect to a Postgres DB
    will differ from Dynamo DB.
    """
    connection_config = get_connection_config_or_error(db, connection_key)

    return validate_and_update_secrets(
        connection_key, connection_config, db, unvalidated_secrets, verify
    )


@router.patch(
    CONNECTION_SECRETS,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[CONNECTION_CREATE_OR_UPDATE])],
    response_model=TestStatusMessage,
)
def patch_connection_config_secrets(
    connection_key: FidesKey,
    *,
    db: Session = Depends(deps.get_db),
    unvalidated_secrets: connection_secrets_schemas,
    verify: Optional[bool] = True,
) -> TestStatusMessage:
    """
    Partially update secrets that will be used to connect to a specified connection_type.

    The specific secrets will be connection-dependent. For example, the components needed to connect to a Postgres DB
    will differ from Dynamo DB.
    """
    connection_config = get_connection_config_or_error(db, connection_key)

    existing_secrets: Optional[Dict[str, Any]] = connection_config.secrets

    # We create the new secrets object by combining the existing secrets with the new secrets.
    patched_secrets = {}
    if existing_secrets:
        patched_secrets = {**existing_secrets}

    patched_secrets = {
        **patched_secrets,
        **unvalidated_secrets,  # type: ignore[dict-item]
    }

    return validate_and_update_secrets(
        connection_key, connection_config, db, patched_secrets, verify  # type: ignore[arg-type]
    )


@router.get(
    CONNECTION_TEST,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[CONNECTION_READ])],
    response_model=TestStatusMessage,
)
def test_connection_config_secrets(
    connection_key: FidesKey,
    *,
    db: Session = Depends(deps.get_db),
) -> TestStatusMessage:
    """
    Endpoint to test a connection at any time using the saved configuration secrets.
    """
    connection_config = get_connection_config_or_error(db, connection_key)
    msg = f"Test completed for ConnectionConfig with key: {connection_key}."
    return connection_status(connection_config, msg, db)



from fides.api.service.connectors.erasure_email_connector import (
    GenericErasureEmailConnector,
)


class AttentiveConnector(GenericErasureEmailConnector):
    """No custom logic here but we want to consider this as a separate connector"""



"""add identity salt table

Revision ID: a249a089f23b
Revises: 784f145ec892
Create Date: 2024-09-04 23:10:42.920719

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op

# revision identifiers, used by Alembic.
revision = "a249a089f23b"
down_revision = "784f145ec892"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "identity_salt",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "encrypted_value",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column("single_row", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.CheckConstraint("single_row", name="identity_salt_single_row_check"),
        sa.UniqueConstraint("single_row", name="identity_salt_single_row_unique"),
    )
    op.create_index(op.f("ix_identity_salt_id"), "identity_salt", ["id"], unique=False)


def downgrade():
    op.drop_index(op.f("ix_identity_salt_id"), table_name="identity_salt")
    op.drop_table("identity_salt")



import os
from enum import Enum as EnumType
from typing import IO, Any, Optional

from loguru import logger as log
from sqlalchemy import Column
from sqlalchemy import Enum as EnumColumn
from sqlalchemy import ForeignKey, String, UniqueConstraint
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session, relationship

from fides.api.db.base_class import Base
from fides.api.models.fides_user import FidesUser  # pylint: disable=unused-import
from fides.api.models.storage import StorageConfig  # pylint: disable=unused-import
from fides.api.schemas.storage.storage import StorageDetails, StorageType
from fides.api.service.storage.s3 import (
    generic_delete_from_s3,
    generic_retrieve_from_s3,
    generic_upload_to_s3,
)
from fides.api.service.storage.util import (
    LOCAL_FIDES_UPLOAD_DIRECTORY,
    get_local_filename,
)


class AttachmentType(str, EnumType):
    """
    Enum for attachment types. Indicates attachment usage.
    """

    internal_use_only = "internal_use_only"
    include_with_access_package = "include_with_access_package"


class AttachmentReferenceType(str, EnumType):
    """
    Enum for attachment reference types. Indicates where attachment is referenced.
    """

    manual_step = "manual_step"
    privacy_request = "privacy_request"
    comment = "comment"


class AttachmentReference(Base):
    """
    Stores information about an Attachment and any other element which may reference that attachment.
    """

    @declared_attr
    def __tablename__(cls) -> str:
        """Overriding base class method to set the table name."""
        return "attachment_reference"

    attachment_id = Column(String, ForeignKey("attachment.id"), nullable=False)
    reference_id = Column(String, nullable=False)
    reference_type = Column(EnumColumn(AttachmentReferenceType), nullable=False)

    __table_args__ = (
        UniqueConstraint(
            "attachment_id", "reference_id", name="_attachment_reference_uc"
        ),
    )

    attachment = relationship(
        "Attachment",
        back_populates="references",
        uselist=False,
    )

    @classmethod
    def create(
        cls, db: Session, *, data: dict[str, Any], check_name: bool = False
    ) -> "AttachmentReference":
        """Creates a new attachment reference record in the database."""
        return super().create(db=db, data=data, check_name=check_name)


class Attachment(Base):
    """
    Stores information about an Attachment.
    """

    user_id = Column(
        String, ForeignKey("fidesuser.id", ondelete="SET NULL"), nullable=True
    )
    file_name = Column(String, nullable=False)
    attachment_type = Column(EnumColumn(AttachmentType), nullable=False)
    storage_key = Column(
        String, ForeignKey("storageconfig.key", ondelete="CASCADE"), nullable=False
    )

    user = relationship(
        "FidesUser",
        lazy="selectin",
        uselist=False,
    )

    references = relationship(
        "AttachmentReference",
        back_populates="attachment",
        cascade="all, delete",
        uselist=True,
    )

    config = relationship(
        "StorageConfig",
        lazy="selectin",
        uselist=False,
    )

    def upload(self, attachment: IO[bytes]) -> None:
        """Uploads an attachment to S3 or local storage."""
        if self.config.type == StorageType.s3:
            bucket_name = f"{self.config.details[StorageDetails.BUCKET.value]}"
            auth_method = self.config.details[StorageDetails.AUTH_METHOD.value]
            generic_upload_to_s3(
                storage_secrets=self.config.secrets,
                bucket_name=bucket_name,
                file_key=self.id,
                document=attachment,
                auth_method=auth_method,
            )
            log.info(f"Uploaded {self.file_name} to S3 bucket {bucket_name}/{self.id}")
            return

        if self.config.type == StorageType.local:
            filename = get_local_filename(self.id)
            with open(filename, "wb") as file:
                file.write(attachment.read())
            return

        raise ValueError(f"Unsupported storage type: {self.config.type}")

    def retrieve_attachment(self) -> Optional[bytes]:
        """Returns the attachment from S3 in bytes form."""
        if self.config.type == StorageType.s3:
            bucket_name = f"{self.config.details[StorageDetails.BUCKET.value]}"
            auth_method = self.config.details[StorageDetails.AUTH_METHOD.value]
            return generic_retrieve_from_s3(
                storage_secrets=self.config.secrets,
                bucket_name=bucket_name,
                file_key=self.id,
                auth_method=auth_method,
            )

        if self.config.type == StorageType.local:
            filename = f"{LOCAL_FIDES_UPLOAD_DIRECTORY}/{self.id}"
            with open(filename, "rb") as file:
                return file.read(), filename

        raise ValueError(f"Unsupported storage type: {self.config.type}")

    def delete_attachment_from_storage(self) -> None:
        """Deletes an attachment from S3 or local storage."""
        if self.config.type == StorageType.s3:
            bucket_name = f"{self.config.details[StorageDetails.BUCKET.value]}"
            auth_method = self.config.details[StorageDetails.AUTH_METHOD.value]
            generic_delete_from_s3(
                storage_secrets=self.config.secrets,
                bucket_name=bucket_name,
                file_key=self.id,
                auth_method=auth_method,
            )
            return

        if self.config.type == StorageType.local:
            filename = f"{LOCAL_FIDES_UPLOAD_DIRECTORY}/{self.id}"
            os.remove(filename)
            return

        raise ValueError(f"Unsupported storage type: {self.config.type}")

    @classmethod
    def create_and_upload(
        cls,
        db: Session,
        *,
        data: dict[str, Any],
        attachment_file: IO[bytes],
        check_name: bool = False,
    ) -> "Attachment":
        """Creates a new attachment record in the database and uploads the attachment to S3."""
        if attachment_file is None:
            raise ValueError("Attachment is required")
        attachment_model = super().create(db=db, data=data, check_name=check_name)

        try:
            attachment_model.upload(attachment_file)
            return attachment_model
        except Exception as e:
            log.error(f"Failed to upload attachment: {e}")
            attachment_model.delete(db)
            raise e

    @classmethod
    def create(
        cls,
        db: Session,
        *,
        data: dict[str, Any],
        check_name: bool = False,
    ) -> "Attachment":
        """Raises Error, provides information for user to create with upload instead."""
        raise NotImplementedError("Please use create_and_upload method for Attachment")

    def delete(self, db: Session) -> None:
        """Deletes an attachment record from the database and deletes the attachment from S3."""
        self.delete_attachment_from_storage()
        super().delete(db=db)



from enum import Enum as EnumType
from typing import TYPE_CHECKING, Any

from sqlalchemy import Column
from sqlalchemy import Enum as EnumColumn
from sqlalchemy import ForeignKey, String, UniqueConstraint
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session, relationship

from fides.api.db.base_class import Base

if TYPE_CHECKING:
    from fides.api.models.attachment import Attachment
    from fides.api.models.fides_user import FidesUser


class CommentType(str, EnumType):
    """
    Enum for comment types. Indicates comment usage.

    - notes are internal comments.
    - reply comments are public and may cause an email or other communciation to be sent
    """

    note = "note"
    reply = "reply"


class CommentReferenceType(str, EnumType):
    """
    Enum for comment reference types. Indicates where the comment is referenced.
    """

    manual_step = "manual_step"
    privacy_request = "privacy_request"


class CommentReference(Base):
    """
    Stores information about a comment and any other element which may reference that comment.
    """

    @declared_attr
    def __tablename__(cls) -> str:
        """Overriding base class method to set the table name."""
        return "comment_reference"

    comment_id = Column(String, ForeignKey("comment.id"), nullable=False)
    reference_id = Column(String, nullable=False)
    reference_type = Column(EnumColumn(CommentReferenceType), nullable=False)

    __table_args__ = (
        UniqueConstraint("comment_id", "reference_id", name="comment_reference_uc"),
    )

    comment = relationship(
        "Comment",
        back_populates="references",
        uselist=False,
    )

    @classmethod
    def create(
        cls, db: Session, *, data: dict[str, Any], check_name: bool = False
    ) -> "CommentReference":
        """Creates a new comment reference record in the database."""
        return super().create(db=db, data=data, check_name=check_name)


class Comment(Base):
    """
    Stores information about a Comment.
    """

    user_id = Column(
        String, ForeignKey("fidesuser.id", ondelete="SET NULL"), nullable=True
    )
    comment_text = Column(String, nullable=False)
    comment_type = Column(EnumColumn(CommentType), nullable=False)

    user = relationship(
        "FidesUser",
        lazy="selectin",
        uselist=False,
    )

    references = relationship(
        "CommentReference",
        back_populates="comment",
        cascade="all, delete",
        uselist=True,
    )

    attachments = relationship(
        "Attachment",
        secondary="attachment_reference",
        primaryjoin="Comment.id == AttachmentReference.reference_id",
        secondaryjoin="Attachment.id == AttachmentReference.attachment_id",
        order_by="Attachment.created_at",
        uselist=True,
    )

    def delete(self, db: Session) -> None:
        """Delete the comment and all associated references."""
        # Delete the comment
        for attachment in self.attachments:
            if len(attachment.references) == 1:
                attachment.delete(db)
        db.delete(self)



from typing import Optional

from loguru import logger

from fides.api.common_exceptions import IdentityVerificationException
from fides.api.util.cache import FidesopsRedis, get_cache
from fides.config import CONFIG


class IdentityVerificationMixin:
    """
    A class housing common identity verification logic for use as a mixin with
    any sqlalchemy model with an ID.
    """

    def _get_identity_verification_cache_key(self) -> str:
        """
        Returns the cache key at which the identity verification code is stored.
        """
        return f"IDENTITY_VERIFICATION_CODE__{self.id}"  # type: ignore

    def _get_identity_verification_attempt_count_cache_key(self) -> str:
        """
        Returns the cache key at which the attempt count for this request's identity
        verification is stored.
        """
        return self._get_identity_verification_cache_key() + "__attempt_count"

    def cache_identity_verification_code(self, value: str) -> None:
        """Cache the generated identity verification code for later comparison."""
        cache: FidesopsRedis = get_cache()
        cache.set_with_autoexpire(
            key=self._get_identity_verification_cache_key(),
            value=value,
            expire_time=CONFIG.redis.identity_verification_code_ttl_seconds,
        )
        cache.set_with_autoexpire(
            key=self._get_identity_verification_attempt_count_cache_key(),
            value=0,
            expire_time=CONFIG.redis.identity_verification_code_ttl_seconds,
        )

    def _increment_verification_code_attempt_count(self) -> None:
        """Cache the generated identity verification code for later comparison."""
        cache: FidesopsRedis = get_cache()
        attempt_count: int = self._get_cached_verification_code_attempt_count()
        cache.set_with_autoexpire(
            key=self._get_identity_verification_attempt_count_cache_key(),
            value=attempt_count + 1,
            expire_time=CONFIG.redis.identity_verification_code_ttl_seconds,
        )

    def get_cached_verification_code(self) -> Optional[str]:
        """Retrieve the generated identity verification code if it exists"""
        cache = get_cache()
        values = cache.get_values([self._get_identity_verification_cache_key()]) or {}
        if not values:
            return None

        return values.get(self._get_identity_verification_cache_key(), None)

    def _get_cached_verification_code_attempt_count(self) -> int:
        """Retrieve the generated identity verification code if it exists"""
        cache = get_cache()
        attempts = cache.get(self._get_identity_verification_attempt_count_cache_key())
        if not attempts:
            attempts = "0"
        return int(attempts)

    def purge_verification_code(self) -> None:
        """Removes any verification codes from the cache so they can no longer be used."""
        logger.debug(
            "Removing cached identity verification code for record with ID: {}",
            self.id,  # type: ignore
        )
        cache = get_cache()
        cache.delete(self._get_identity_verification_cache_key())
        cache.delete(self._get_identity_verification_attempt_count_cache_key())

    def _verify_identity(self, provided_code: Optional[str] = None) -> None:
        """Verify the identification code supplied by the user."""
        code: Optional[str] = self.get_cached_verification_code()
        if not code:
            raise IdentityVerificationException(
                f"Identification code expired for {self.id}."  # type: ignore
            )

        attempt_count: int = self._get_cached_verification_code_attempt_count()
        if attempt_count >= CONFIG.security.identity_verification_attempt_limit:
            logger.debug(
                "Failed identity verification attempt limit exceeded for record with ID: {}",
                self.id,  # type: ignore
            )
            # When the attempt_count we can remove the verification code entirely
            # from the cache to ensure it can never be used again.
            self.purge_verification_code()
            raise PermissionError(f"Attempt limit hit for '{self.id}'")  # type: ignore

        if code != provided_code:
            self._increment_verification_code_attempt_count()
            raise PermissionError(f"Incorrect identification code for '{self.id}'")  # type: ignore



import os

LOCAL_FIDES_UPLOAD_DIRECTORY = "fides_uploads"


def get_local_filename(file_key: str) -> str:
    """Verifies that the local storage directory exists"""
    if not os.path.exists(LOCAL_FIDES_UPLOAD_DIRECTORY):
        os.makedirs(LOCAL_FIDES_UPLOAD_DIRECTORY)
    return f"{LOCAL_FIDES_UPLOAD_DIRECTORY}/{file_key}"
